import pytest

# This file can be used to define fixtures and other test configuration
# that will be available to all test files

# Enable asyncio support for pytest
pytest_plugins = ["pytest_asyncio"]



import os
import pytest
import tempfile
import hashlib


from src.text_editor.server import TextEditorServer, calculate_id, generate_diff_preview
from mcp.server.fastmcp import FastMCP


class TestTextEditorServer:
    @pytest.fixture
    def server(self, monkeypatch):
        """Create a TextEditorServer instance for testing."""
        monkeypatch.setenv("PYTHON_VENV", "python")
        server = TextEditorServer()
        server.max_select_lines = 200
        return server

    @pytest.fixture
    def temp_file(self):
        """Create a temporary file for testing."""
        content = "Line 1\nLine 2\nLine 3\nLine 4\nLine 5\n"
        with tempfile.NamedTemporaryFile(mode="w+", delete=False) as f:
            f.write(content)
            temp_path = f.name
        yield temp_path
        if os.path.exists(temp_path):
            os.unlink(temp_path)

    @pytest.fixture
    def empty_temp_file(self):
        """Create an empty temporary file for testing."""
        with tempfile.NamedTemporaryFile(mode="w+", delete=False) as f:
            temp_path = f.name
        yield temp_path
        if os.path.exists(temp_path):
            os.unlink(temp_path)

    @pytest.fixture
    def server_with_protected_paths(self):
        """Create a TextEditorServer instance with protected paths configuration."""
        server = TextEditorServer()
        server.max_select_lines = 200
        # Define protected paths for testing
        server.protected_paths = ["*.env", "/etc/passwd", "/home/secret-file.txt"]
        return server

    def get_tool_fn(self, server, tool_name):
        """Helper to get the tool function from the server."""
        tools_dict = server.mcp._tool_manager._tools
        return tools_dict[tool_name].fn

    @pytest.mark.asyncio
    async def test_set_file_valid(self, server, temp_file):
        """Test setting a valid file path."""
        set_file_fn = self.get_tool_fn(server, "set_file")
        result = await set_file_fn(temp_file)
        assert "File set to:" in result
        assert temp_file in result
        assert server.current_file_path == temp_file

    @pytest.mark.asyncio
    async def test_set_file_protected_path_exact_match(
        self, server_with_protected_paths
    ):
        """Test setting a file path that exactly matches a protected path."""
        set_file_fn = self.get_tool_fn(server_with_protected_paths, "set_file")
        result = await set_file_fn("/etc/passwd")
        assert "Error: Access to '/etc/passwd' is denied" in result
        assert server_with_protected_paths.current_file_path is None

    @pytest.mark.asyncio
    async def test_set_file_protected_path_wildcard_match(
        self, server_with_protected_paths, monkeypatch
    ):
        """Test setting a file path that matches a wildcard protected path pattern."""
        # Create a temporary .env file for testing
        with tempfile.NamedTemporaryFile(mode="w+", suffix=".env", delete=False) as f:
            f.write("API_KEY=test_key\n")
            env_file_path = f.name

        # Mock os.path.isfile to return True for our temp file
        def mock_isfile(path):
            return path == env_file_path

        monkeypatch.setattr(os.path, "isfile", mock_isfile)

        try:
            set_file_fn = self.get_tool_fn(server_with_protected_paths, "set_file")
            result = await set_file_fn(env_file_path)
            assert "Error: Access to '" in result
            assert (
                "is denied due to PROTECTED_PATHS configuration (matches pattern '*.env')"
                in result
            )
            assert server_with_protected_paths.current_file_path is None
        finally:
            if os.path.exists(env_file_path):
                os.unlink(env_file_path)

    @pytest.mark.asyncio
    async def test_set_file_protected_path_glob_match(self, monkeypatch):
        """Test setting a file path that matches a more complex glob pattern."""
        # Create a server with different glob patterns
        server = TextEditorServer()
        server.protected_paths = [".env*", "config*.json", "*keys.txt"]

        # Create a temporary .env.local file for testing
        with tempfile.NamedTemporaryFile(
            mode="w+", prefix=".env", suffix=".local", delete=False
        ) as f:
            f.write("API_KEY=test_key\n")
            env_local_path = f.name

        # Create a temporary config-dev.json file for testing
        with tempfile.NamedTemporaryFile(
            mode="w+", prefix="config-", suffix=".json", delete=False
        ) as f:
            f.write('{"debug": true}\n')
            config_path = f.name

        # Create a custom filename that will definitely match our pattern
        keys_file_path = os.path.join(tempfile.gettempdir(), "api-keys.txt")
        with open(keys_file_path, "w") as f:
            f.write("secret_key=abc123\n")
        secret_path = keys_file_path

        # Mock os.path.isfile to return True for our test files
        def mock_isfile(path):
            return path in [env_local_path, config_path, secret_path]

        monkeypatch.setattr(os.path, "isfile", mock_isfile)

        try:
            set_file_fn = self.get_tool_fn(server, "set_file")

            # Test .env* pattern
            result = await set_file_fn(env_local_path)
            assert "Error: Access to '" in result
            assert (
                "is denied due to PROTECTED_PATHS configuration (matches pattern '.env*'"
                in result
            )
            assert server.current_file_path is None

            # Test config*.json pattern
            result = await set_file_fn(config_path)
            assert "Error: Access to '" in result
            assert (
                "is denied due to PROTECTED_PATHS configuration (matches pattern 'config*.json'"
                in result
            )
            assert server.current_file_path is None

            # Test *keys.txt pattern
            result = await set_file_fn(secret_path)
            assert "Error: Access to '" in result
            assert (
                "is denied due to PROTECTED_PATHS configuration (matches pattern '*keys.txt'"
                in result
            )
            assert server.current_file_path is None
        finally:
            # Clean up temp files
            for path in [env_local_path, config_path, secret_path]:
                if os.path.exists(path):
                    os.unlink(path)

    @pytest.mark.asyncio
    async def test_set_file_non_protected_path(
        self, server_with_protected_paths, temp_file
    ):
        """Test setting a file path that does not match any protected paths."""
        set_file_fn = self.get_tool_fn(server_with_protected_paths, "set_file")
        result = await set_file_fn(temp_file)
        assert "File set to:" in result
        assert temp_file in result
        assert server_with_protected_paths.current_file_path == temp_file

    @pytest.mark.asyncio
    async def test_set_file_invalid(self, server):
        """Test setting a non-existent file path."""
        set_file_fn = self.get_tool_fn(server, "set_file")
        non_existent_path = "/path/to/nonexistent/file.txt"
        result = await set_file_fn(non_existent_path)
        assert "Error: File not found" in result
        assert server.current_file_path is None

    @pytest.mark.asyncio
    async def test_read_no_file_set(self, server):
        """Test getting text when no file is set."""
        read_fn = self.get_tool_fn(server, "read")
        result = await read_fn(1, 10)
        assert "error" in result
        assert "No file path is set" in result["error"]

    @pytest.mark.asyncio
    async def test_read_entire_file(self, server, temp_file):
        """Test getting the entire content of a file."""
        set_file_fn = self.get_tool_fn(server, "set_file")
        await set_file_fn(temp_file)
        read_fn = self.get_tool_fn(server, "read")
        result = await read_fn(1, 5)
        assert "lines" in result

    async def test_read_line_range(self, server, temp_file):
        """Test getting a specific range of lines from a file."""
        set_file_fn = self.get_tool_fn(server, "set_file")
        await set_file_fn(temp_file)
        read_fn = self.get_tool_fn(server, "read")
        result = await read_fn(2, 4)
        assert "lines" in result
        select_fn = self.get_tool_fn(server, "select")
        select_result = await select_fn(2, 4)
        assert "status" in select_result
        assert "id" in select_result
        expected_id = calculate_id("Line 2\nLine 3\nLine 4\n", 2, 4)
        assert expected_id == select_result["id"]

    @pytest.mark.asyncio
    async def test_read_only_end_line(self, server, temp_file):
        """Test getting text with only end line specified."""
        set_file_fn = self.get_tool_fn(server, "set_file")
        await set_file_fn(temp_file)
        read_fn = self.get_tool_fn(server, "read")
        result = await read_fn(1, 2)
        assert "lines" in result
        select_fn = self.get_tool_fn(server, "select")
        select_result = await select_fn(1, 2)
        expected_id = calculate_id("Line 1\nLine 2\n", 1, 2)
        assert expected_id == select_result["id"]

    @pytest.mark.asyncio
    async def test_read_invalid_range(self, server, temp_file):
        """Test getting text with an invalid line range."""
        set_file_fn = self.get_tool_fn(server, "set_file")
        await set_file_fn(temp_file)
        read_fn = self.get_tool_fn(server, "read")
        result = await read_fn(4, 2)
        assert "error" in result
        # Updated assertion to match actual error message format in server.py
        assert "start=4 cannot be greater than end=2" in result["error"]
        result = await read_fn(0, 3)
        assert "error" in result
        assert "start must be at least 1" in result["error"]

    def test_calculate_id_function(self):
        """Test the calculate_id function directly."""
        text = "Some test content"
        id_no_range = calculate_id(text)
        expected = hashlib.sha256(text.encode()).hexdigest()[:2]
        assert id_no_range == expected
        id_with_range = calculate_id(text, 1, 3)
        assert id_with_range.startswith("L1-3-")
        assert id_with_range.endswith(expected)

    @pytest.mark.asyncio
    async def test_read_large_file(self, server):
        """Test getting text from a file larger than MAX_SELECT_LINES lines."""
        more_than_max_lines = server.max_select_lines + 10
        with tempfile.NamedTemporaryFile(mode="w+", delete=False) as f:
            for i in range(more_than_max_lines):
                f.write(f"Line {i + 1}\n")
            large_file_path = f.name
        try:
            set_file_fn = self.get_tool_fn(server, "set_file")
            await set_file_fn(large_file_path)
            read_fn = self.get_tool_fn(server, "read")
            result = await read_fn(1, more_than_max_lines)
            assert "lines" in result
            select_fn = self.get_tool_fn(server, "select")
            result = await select_fn(1, more_than_max_lines)
            assert "error" in result
            assert (
                f"Cannot select more than {server.max_select_lines} lines at once"
                in result["error"]
            )
            result = await select_fn(5, 15)
            assert "status" in result
            assert "id" in result
            result = await read_fn(5, server.max_select_lines + 10)
            assert "lines" in result
        finally:
            if os.path.exists(large_file_path):
                os.unlink(large_file_path)

    @pytest.mark.asyncio
    async def test_new_file(self, server, empty_temp_file):
        """Test new_file functionality."""
        set_file_fn = self.get_tool_fn(server, "set_file")
        await set_file_fn(empty_temp_file)
        new_file_fn = self.get_tool_fn(server, "new_file")
        result = await new_file_fn(empty_temp_file)
        assert result["status"] == "success"
        assert "id" in result
        result = await new_file_fn(empty_temp_file)
        assert "error" in result

    @pytest.mark.asyncio
    async def test_delete_file(self, server):
        """Test delete_file tool."""
        with tempfile.NamedTemporaryFile(mode="w+", delete=False) as f:
            f.write("Test content to delete")
            temp_path = f.name
        try:
            delete_file_fn = self.get_tool_fn(server, "delete_file")
            result = await delete_file_fn()
            assert "error" in result
            assert "No file path is set" in result["error"]
            set_file_fn = self.get_tool_fn(server, "set_file")
            await set_file_fn(temp_path)
            result = await delete_file_fn()
            assert result["status"] == "success"
            assert "successfully deleted" in result["message"]
            assert temp_path in result["message"]
            assert not os.path.exists(temp_path)
            assert server.current_file_path is None
            result = await set_file_fn(temp_path)
            assert "Error: File not found" in result
            assert server.current_file_path is None
        finally:
            if os.path.exists(temp_path):
                os.unlink(temp_path)

    @pytest.mark.asyncio
    async def test_delete_file_permission_error(self, server, monkeypatch):
        """Test delete_file with permission error."""
        with tempfile.NamedTemporaryFile(mode="w+", delete=False) as f:
            f.write("Test content")
            temp_path = f.name
        try:
            set_file_fn = self.get_tool_fn(server, "set_file")
            await set_file_fn(temp_path)

            def mock_remove(path):
                raise PermissionError("Permission denied")

            monkeypatch.setattr(os, "remove", mock_remove)
            delete_file_fn = self.get_tool_fn(server, "delete_file")
            result = await delete_file_fn()
            assert "error" in result
            assert "Permission denied" in result["error"]
            assert server.current_file_path == temp_path
        finally:
            monkeypatch.undo()
            if os.path.exists(temp_path):
                os.unlink(temp_path)

    @pytest.mark.asyncio
    async def test_find_line_no_file_set(self, server):
        """Test find_line with no file set."""
        find_line_fn = self.get_tool_fn(server, "find_line")
        result = await find_line_fn(search_text="Line")
        assert "error" in result
        assert "No file path is set" in result["error"]

    @pytest.mark.asyncio
    async def test_find_line_basic(self, server, temp_file):
        """Test basic find_line functionality."""
        set_file_fn = self.get_tool_fn(server, "set_file")
        await set_file_fn(temp_file)
        find_line_fn = self.get_tool_fn(server, "find_line")
        result = await find_line_fn(search_text="Line")
        assert "status" in result
        assert result["status"] == "success"
        assert "matches" in result
        assert "total_matches" in result
        assert result["total_matches"] == 5
        for match in result["matches"]:
            # Each match is a list with [line_number, line_text]
            assert len(match) == 2
            assert isinstance(match[0], int)  # line number
            assert isinstance(match[1], str)  # line text
            assert f"Line {match[0]}" in match[1]
        line_numbers = [match[0] for match in result["matches"]]
        assert line_numbers == [1, 2, 3, 4, 5]

    @pytest.mark.asyncio
    async def test_find_line_specific_match(self, server, temp_file):
        """Test find_line with a specific search term."""
        set_file_fn = self.get_tool_fn(server, "set_file")
        await set_file_fn(temp_file)
        find_line_fn = self.get_tool_fn(server, "find_line")
        result = await find_line_fn(search_text="Line 3")
        assert result["status"] == "success"
        assert result["total_matches"] == 1
        assert len(result["matches"]) == 1
        assert result["matches"][0][0] == 3  # First element is the line number
        assert "Line 3" in result["matches"][0][1]  # Second element is the line text

    @pytest.mark.asyncio
    async def test_find_line_no_matches(self, server, temp_file):
        """Test find_line with a search term that doesn't exist."""
        set_file_fn = self.get_tool_fn(server, "set_file")
        await set_file_fn(temp_file)
        find_line_fn = self.get_tool_fn(server, "find_line")
        result = await find_line_fn(search_text="NonExistentTerm")
        assert result["status"] == "success"
        assert result["total_matches"] == 0
        assert len(result["matches"]) == 0

    @pytest.mark.asyncio
    async def test_skim_no_file_set(self, server):
        """Test skim with no file set."""
        skim_fn = self.get_tool_fn(server, "skim")
        result = await skim_fn()
        assert "error" in result
        assert "No file path is set" in result["error"]

    @pytest.mark.asyncio
    async def test_skim_basic(self, server, temp_file):
        """Test basic skim functionality."""
        set_file_fn = self.get_tool_fn(server, "set_file")
        await set_file_fn(temp_file)
        skim_fn = self.get_tool_fn(server, "skim")
        result = await skim_fn()
        assert "lines" in result
        assert "total_lines" in result
        assert "max_select_lines" in result
        assert result["total_lines"] == 5
        assert result["max_select_lines"] == server.max_select_lines
        assert len(result["lines"]) == 5
        for i, line_data in enumerate(result["lines"], 1):
            assert line_data[0] == i  # Check line number
            assert line_data[1] == f"Line {i}"  # Check line content

    @pytest.mark.asyncio
    async def test_overwrite_no_selection(self, server, temp_file):
        """Test overwrite when no selection has been made."""
        set_file_fn = self.get_tool_fn(server, "set_file")
        await set_file_fn(temp_file)
        overwrite_fn = self.get_tool_fn(server, "overwrite")
        result = await overwrite_fn(new_lines={"lines": ["New content"]})
        assert "error" in result
        assert "No selection has been made" in result["error"]

    @pytest.mark.asyncio
    async def test_find_line_file_read_error(self, server, temp_file, monkeypatch):
        """Test find_line with a file read error."""
        set_file_fn = self.get_tool_fn(server, "set_file")
        await set_file_fn(temp_file)

        def mock_open(*args, **kwargs):
            raise IOError("Mock file read error")

        monkeypatch.setattr("builtins.open", mock_open)
        find_line_fn = self.get_tool_fn(server, "find_line")
        result = await find_line_fn(search_text="Line")
        assert "error" in result
        assert "Error searching file" in result["error"]
        assert "Mock file read error" in result["error"]

    @pytest.mark.asyncio
    async def test_overwrite_no_file_set(self, server):
        """Test overwrite when no file is set."""
        overwrite_fn = self.get_tool_fn(server, "overwrite")
        result = await overwrite_fn(new_lines={"lines": ["New content"]})
        assert "error" in result
        assert "No file path is set" in result["error"]

    @pytest.mark.asyncio
    async def test_overwrite_basic(self, server, temp_file):
        """Test basic overwrite functionality."""
        set_file_fn = self.get_tool_fn(server, "set_file")
        await set_file_fn(temp_file)
        select_fn = self.get_tool_fn(server, "select")
        select_result = await select_fn(2, 4)
        assert select_result["status"] == "success"
        assert "id" in select_result
        overwrite_fn = self.get_tool_fn(server, "overwrite")
        new_lines = {"lines": ["New Line 2", "New Line 3", "New Line 4"]}
        result = await overwrite_fn(new_lines=new_lines)
        assert "status" in result
        assert result["status"] == "preview"
        assert "Changes ready to apply" in result["message"]
        confirm_fn = self.get_tool_fn(server, "confirm")
        confirm_result = await confirm_fn()
        assert confirm_result["status"] == "success"
        assert "Changes applied successfully" in confirm_result["message"]
        with open(temp_file, "r") as f:
            file_content = f.read()
        expected_content = "Line 1\nNew Line 2\nNew Line 3\nNew Line 4\nLine 5\n"
        assert file_content == expected_content

    @pytest.mark.asyncio
    async def test_overwrite_cancel(self, server, temp_file):
        """Test overwrite with cancel operation."""
        # Set up initial state
        set_file_fn = self.get_tool_fn(server, "set_file")
        await set_file_fn(temp_file)
        select_fn = self.get_tool_fn(server, "select")
        select_result = await select_fn(2, 4)
        assert select_result["status"] == "success"
        assert "id" in select_result

        # Create overwrite preview
        overwrite_fn = self.get_tool_fn(server, "overwrite")
        new_lines = {"lines": ["New Line 2", "New Line 3", "New Line 4"]}
        result = await overwrite_fn(new_lines=new_lines)
        assert "status" in result
        assert result["status"] == "preview"
        assert "Changes ready to apply" in result["message"]

        # Get original content to verify it remains unchanged
        with open(temp_file, "r") as f:
            original_content = f.read()

        # Cancel the changes
        cancel_fn = self.get_tool_fn(server, "cancel")
        cancel_result = await cancel_fn()
        assert cancel_result["status"] == "success"
        assert "Action cancelled" in cancel_result["message"]

        # Verify the file content is unchanged
        with open(temp_file, "r") as f:
            file_content = f.read()
        assert file_content == original_content

        # Verify that selected lines are still available
        assert server.selected_start == 2
        assert server.selected_end == 4
        assert server.selected_id is not None
        assert server.pending_modified_lines is None
        assert server.pending_diff is None

    @pytest.mark.asyncio
    async def test_select_invalid_range(self, server, temp_file):
        """Test select with invalid line ranges."""
        set_file_fn = self.get_tool_fn(server, "set_file")
        await set_file_fn(temp_file)
        select_fn = self.get_tool_fn(server, "select")
        result = await select_fn(start=0, end=2)
        assert "error" in result
        assert "start must be at least 1" in result["error"]
        result = await select_fn(start=1, end=10)
        assert "end" in result
        assert result["end"] == 5
        result = await select_fn(start=4, end=2)
        assert "error" in result
        assert "start cannot be greater than end" in result["error"]

    @pytest.mark.asyncio
    async def test_overwrite_id_verification_failed(self, server, temp_file):
        """Test overwrite with incorrect ID (content verification failure)."""
        set_file_fn = self.get_tool_fn(server, "set_file")
        await set_file_fn(temp_file)
        select_fn = self.get_tool_fn(server, "select")
        select_result = await select_fn(2, 3)
        with open(temp_file, "w") as f:
            f.write(
                "Modified Line 1\nModified Line 2\nModified Line 3\nModified Line 4\nModified Line 5\n"
            )
        overwrite_fn = self.get_tool_fn(server, "overwrite")
        result = await overwrite_fn(new_lines={"lines": ["New content"]})
        assert "error" in result
        assert "id verification failed" in result["error"]

    @pytest.mark.asyncio
    async def test_overwrite_different_line_count(self, server, temp_file):
        """Test overwrite with different line count (more or fewer lines)."""
        set_file_fn = self.get_tool_fn(server, "set_file")
        await set_file_fn(temp_file)
        select_fn = self.get_tool_fn(server, "select")
        select_result = await select_fn(2, 3)
        assert select_result["status"] == "success"
        overwrite_fn = self.get_tool_fn(server, "overwrite")
        new_lines = {"lines": ["New Line 2", "Extra Line", "New Line 3"]}
        result = await overwrite_fn(new_lines=new_lines)
        assert result["status"] == "preview"
        confirm_fn = self.get_tool_fn(server, "confirm")
        confirm_result = await confirm_fn()
        assert confirm_result["status"] == "success"
        with open(temp_file, "r") as f:
            file_content = f.read()
        expected_content = (
            "Line 1\nNew Line 2\nExtra Line\nNew Line 3\nLine 4\nLine 5\n"
        )
        assert file_content == expected_content
        select_result = await select_fn(1, 6)
        assert select_result["status"] == "success"
        new_content = "Single Line\n"
        result = await overwrite_fn(new_lines={"lines": ["Single Line"]})
        assert result["status"] == "preview"
        confirm_result = await confirm_fn()
        assert confirm_result["status"] == "success"
        with open(temp_file, "r") as f:
            file_content = f.read()
        assert file_content == "Single Line\n"

    @pytest.mark.asyncio
    async def test_overwrite_empty_text(self, server, temp_file):
        """Test overwrite with empty text (effectively removing lines)."""
        set_file_fn = self.get_tool_fn(server, "set_file")
        await set_file_fn(temp_file)
        select_fn = self.get_tool_fn(server, "select")
        select_result = await select_fn(2, 3)
        assert select_result["status"] == "success"
        overwrite_fn = self.get_tool_fn(server, "overwrite")
        result = await overwrite_fn(new_lines={"lines": []})
        assert result["status"] == "preview"
        confirm_fn = self.get_tool_fn(server, "confirm")
        confirm_result = await confirm_fn()
        assert confirm_result["status"] == "success"
        with open(temp_file, "r") as f:
            file_content = f.read()
        expected_content = "Line 1\nLine 4\nLine 5\n"
        assert file_content == expected_content

    @pytest.mark.asyncio
    async def test_select_max_lines_exceeded(self, server, temp_file):
        """Test select with a range exceeding max_select_lines."""
        set_file_fn = self.get_tool_fn(server, "set_file")
        await set_file_fn(temp_file)
        more_than_max_lines = server.max_select_lines + 10
        with tempfile.NamedTemporaryFile(mode="w+", delete=False) as f:
            for i in range(more_than_max_lines):
                f.write(f"Line {i + 1}\n")
            large_file_path = f.name
        try:
            await set_file_fn(large_file_path)
            select_fn = self.get_tool_fn(server, "select")
            result = await select_fn(start=1, end=server.max_select_lines + 1)
            assert "error" in result
            assert (
                f"Cannot select more than {server.max_select_lines} lines at once"
                in result["error"]
            )
        finally:
            if os.path.exists(large_file_path):
                os.unlink(large_file_path)

    @pytest.mark.asyncio
    async def test_overwrite_file_read_error(self, server, temp_file, monkeypatch):
        """Test overwrite with file read error."""
        set_file_fn = self.get_tool_fn(server, "set_file")
        await set_file_fn(temp_file)
        select_fn = self.get_tool_fn(server, "select")
        select_result = await select_fn(2, 3)
        assert select_result["status"] == "success"
        original_open = open

        def mock_open_read(*args, **kwargs):
            if args[1] == "r":
                raise IOError("Mock file read error")
            return original_open(*args, **kwargs)

        monkeypatch.setattr("builtins.open", mock_open_read)
        overwrite_fn = self.get_tool_fn(server, "overwrite")
        result = await overwrite_fn(new_lines={"lines": ["New content"]})
        assert "error" in result
        assert "Error reading file" in result["error"]
        assert "Mock file read error" in result["error"]

    @pytest.mark.asyncio
    async def test_overwrite_file_write_error(self, server, temp_file, monkeypatch):
        """Test overwrite with file write error."""
        set_file_fn = self.get_tool_fn(server, "set_file")
        await set_file_fn(temp_file)
        select_fn = self.get_tool_fn(server, "select")
        select_result = await select_fn(2, 3)
        assert select_result["status"] == "success"
        original_open = open
        open_calls = [0]

        def mock_open_write(*args, **kwargs):
            if args[1] == "w":
                raise IOError("Mock file write error")
            return original_open(*args, **kwargs)

        monkeypatch.setattr("builtins.open", mock_open_write)
        overwrite_fn = self.get_tool_fn(server, "overwrite")
        result = await overwrite_fn(new_lines={"lines": ["New content"]})
        assert "status" in result
        assert result["status"] == "preview"
        confirm_fn = self.get_tool_fn(server, "confirm")
        confirm_result = await confirm_fn()
        assert "error" in confirm_result
        assert "Error writing to file" in confirm_result["error"]
        assert "Mock file write error" in confirm_result["error"]

    @pytest.mark.asyncio
    async def test_overwrite_newline_handling(self, server):
        """Test newline handling in overwrite (appends newline when needed)."""
        with tempfile.NamedTemporaryFile(mode="w+", delete=False) as f:
            f.write("Line 1\nLine 2\nLine 3")
            temp_path = f.name
        try:
            set_file_fn = self.get_tool_fn(server, "set_file")
            await set_file_fn(temp_path)
            select_fn = self.get_tool_fn(server, "select")
            select_result = await select_fn(2, 2)
            assert select_result["status"] == "success"
            overwrite_fn = self.get_tool_fn(server, "overwrite")
            result = await overwrite_fn(new_lines={"lines": ["New Line 2"]})
            assert result["status"] == "preview"
            confirm_fn = self.get_tool_fn(server, "confirm")
            confirm_result = await confirm_fn()
            assert confirm_result["status"] == "success"
            with open(temp_path, "r") as f:
                file_content = f.read()
            expected_content = "Line 1\nNew Line 2\nLine 3"
            assert file_content == expected_content
        finally:
            if os.path.exists(temp_path):
                os.unlink(temp_path)

    @pytest.mark.asyncio
    async def test_overwrite_python_syntax_check_success(self, server):
        """Test Python syntax checking in overwrite succeeds with valid Python code."""
        valid_python_content = (
            "def hello():\n    print('Hello, world!')\n\nresult = hello()\n"
        )
        with tempfile.NamedTemporaryFile(mode="w+", suffix=".py", delete=False) as f:
            f.write(valid_python_content)
            py_file_path = f.name
        try:
            set_file_fn = self.get_tool_fn(server, "set_file")
            await set_file_fn(py_file_path)
            select_fn = self.get_tool_fn(server, "select")
            select_result = await select_fn(1, 4)
            assert select_result["status"] == "success"
            overwrite_fn = self.get_tool_fn(server, "overwrite")
            new_content = {
                "lines": [
                    "def greeting(name):",
                    "    return f'Hello, {name}!'",
                    "",
                    "result = greeting('World')",
                ]
            }
            result = await overwrite_fn(new_lines=new_content)
            assert result["status"] == "preview"
            confirm_fn = self.get_tool_fn(server, "confirm")
            confirm_result = await confirm_fn()
            assert confirm_result["status"] == "success"
            assert "Changes applied successfully" in confirm_result["message"]
            with open(py_file_path, "r") as f:
                file_content = f.read()
            expected_content = "def greeting(name):\n    return f'Hello, {name}!'\n\nresult = greeting('World')\n"
            assert file_content == expected_content
        finally:
            if os.path.exists(py_file_path):
                os.unlink(py_file_path)

    @pytest.mark.asyncio
    async def test_overwrite_python_syntax_check_failure(self, server):
        """Test Python syntax checking in overwrite fails with invalid Python code."""
        valid_python_content = (
            "def hello():\n    print('Hello, world!')\n\nresult = hello()\n"
        )
        with tempfile.NamedTemporaryFile(mode="w+", suffix=".py", delete=False) as f:
            f.write(valid_python_content)
            py_file_path = f.name
        try:
            set_file_fn = self.get_tool_fn(server, "set_file")
            await set_file_fn(py_file_path)
            select_fn = self.get_tool_fn(server, "select")
            select_result = await select_fn(1, 4)
            assert select_result["status"] == "success"
            overwrite_fn = self.get_tool_fn(server, "overwrite")
            invalid_python = {
                "lines": [
                    "def broken_function(:",
                    "    print('Missing parenthesis'",
                    "",
                    "result = broken_function()",
                ]
            }
            result = await overwrite_fn(new_lines=invalid_python)
            assert "error" in result
            assert "Python syntax error:" in result["error"]
            with open(py_file_path, "r") as f:
                file_content = f.read()
            assert file_content == valid_python_content
        finally:
            if os.path.exists(py_file_path):
                os.unlink(py_file_path)

    @pytest.mark.asyncio
    async def test_overwrite_javascript_syntax_check_success(self, server, monkeypatch):
        """Test JavaScript syntax checking in overwrite succeeds with valid JS code."""
        valid_js_content = "function hello() {\n  return 'Hello, world!';\n}\n\nconst result = hello();\n"
        with tempfile.NamedTemporaryFile(mode="w+", suffix=".js", delete=False) as f:
            f.write(valid_js_content)
            js_file_path = f.name

        def mock_subprocess_run(*args, **kwargs):
            class MockCompletedProcess:
                def __init__(self):
                    self.returncode = 0
                    self.stderr = ""
                    self.stdout = ""

            return MockCompletedProcess()

        monkeypatch.setattr("subprocess.run", mock_subprocess_run)
        try:
            set_file_fn = self.get_tool_fn(server, "set_file")
            await set_file_fn(js_file_path)
            select_fn = self.get_tool_fn(server, "select")
            select_result = await select_fn(1, 5)
            assert select_result["status"] == "success"
            overwrite_fn = self.get_tool_fn(server, "overwrite")
            new_lines = {
                "lines": [
                    "function greeting(name) {",
                    "  return `Hello, ${name}!`;",
                    "}",
                    "",
                    "const result = greeting('World');",
                ]
            }
            result = await overwrite_fn(new_lines=new_lines)
            assert result["status"] == "preview"
            confirm_fn = self.get_tool_fn(server, "confirm")
            confirm_result = await confirm_fn()
            assert confirm_result["status"] == "success"
            assert "Changes applied successfully" in confirm_result["message"]
            with open(js_file_path, "r") as f:
                file_content = f.read()
            expected_content = "function greeting(name) {\n  return `Hello, ${name}!`;\n}\n\nconst result = greeting('World');\n"
            assert file_content == expected_content
        finally:
            if os.path.exists(js_file_path):
                os.unlink(js_file_path)

    @pytest.mark.asyncio
    async def test_overwrite_javascript_syntax_check_failure(self, server, monkeypatch):
        """Test JavaScript syntax checking in overwrite fails with invalid JS code."""
        valid_js_content = "function hello() {\n  return 'Hello, world!';\n}\n\nconst result = hello();\n"
        with tempfile.NamedTemporaryFile(mode="w+", suffix=".js", delete=False) as f:
            f.write(valid_js_content)
            js_file_path = f.name

        def mock_subprocess_run(*args, **kwargs):
            class MockCompletedProcess:
                def __init__(self):
                    self.returncode = 1
                    self.stderr = "SyntaxError: Unexpected token (1:19)"
                    self.stdout = ""

            return MockCompletedProcess()

        monkeypatch.setattr("subprocess.run", mock_subprocess_run)
        try:
            set_file_fn = self.get_tool_fn(server, "set_file")
            await set_file_fn(js_file_path)
            select_fn = self.get_tool_fn(server, "select")
            select_result = await select_fn(1, 5)
            overwrite_fn = self.get_tool_fn(server, "overwrite")
            invalid_js = {
                "lines": [
                    "function broken() {",
                    "  return 'Missing closing bracket;",
                    "}",
                    "",
                    "const result = broken();",
                ]
            }
            result = await overwrite_fn(new_lines=invalid_js)
            assert "error" in result
            assert "JavaScript syntax error:" in result["error"]
            with open(js_file_path, "r") as f:
                file_content = f.read()
            assert file_content == valid_js_content
        finally:
            if os.path.exists(js_file_path):
                os.unlink(js_file_path)

    @pytest.mark.asyncio
    async def test_overwrite_jsx_syntax_check_success(self, server, monkeypatch):
        """Test JSX syntax checking in overwrite succeeds with valid React/JSX code."""
        valid_jsx_content = "import React from 'react';\n\nfunction HelloWorld() {\n  return <div>Hello, world!</div>;\n}\n\nexport default HelloWorld;\n"
        with tempfile.NamedTemporaryFile(mode="w+", suffix=".jsx", delete=False) as f:
            f.write(valid_jsx_content)
            jsx_file_path = f.name

        def mock_subprocess_run(*args, **kwargs):
            class MockCompletedProcess:
                def __init__(self):
                    self.returncode = 0
                    self.stderr = ""
                    self.stdout = ""

            return MockCompletedProcess()

        monkeypatch.setattr("subprocess.run", mock_subprocess_run)
        try:
            set_file_fn = self.get_tool_fn(server, "set_file")
            await set_file_fn(jsx_file_path)
            select_fn = self.get_tool_fn(server, "select")
            select_result = await select_fn(1, 7)
            assert select_result["status"] == "success"
            overwrite_fn = self.get_tool_fn(server, "overwrite")
            new_jsx_content = {
                "lines": [
                    "import React from 'react';",
                    "",
                    "function Greeting({ name }) {",
                    "  return <div>Hello, {name}!</div>;",
                    "}",
                    "",
                    "export default Greeting;",
                ]
            }
            result = await overwrite_fn(new_lines=new_jsx_content)
            assert result["status"] == "preview"
            confirm_fn = self.get_tool_fn(server, "confirm")
            confirm_result = await confirm_fn()
            assert confirm_result["status"] == "success"
            assert "Changes applied successfully" in confirm_result["message"]
            with open(jsx_file_path, "r") as f:
                file_content = f.read()
            expected_content = "import React from 'react';\n\nfunction Greeting({ name }) {\n  return <div>Hello, {name}!</div>;\n}\n\nexport default Greeting;\n"
            assert file_content == expected_content
        finally:
            if os.path.exists(jsx_file_path):
                os.unlink(jsx_file_path)

    @pytest.mark.asyncio
    async def test_overwrite_jsx_syntax_check_failure(self, server, monkeypatch):
        """Test JSX syntax checking in overwrite fails with invalid React/JSX code."""
        valid_jsx_content = "import React from 'react';\n\nfunction HelloWorld() {\n  return <div>Hello, world!</div>;\n}\n\nexport default HelloWorld;\n"
        with tempfile.NamedTemporaryFile(mode="w+", suffix=".jsx", delete=False) as f:
            f.write(valid_jsx_content)
            jsx_file_path = f.name

        def mock_subprocess_run(*args, **kwargs):
            class MockCompletedProcess:
                def __init__(self):
                    self.returncode = 1
                    self.stderr = "SyntaxError: Unexpected token (4:10)"
                    self.stdout = ""

            return MockCompletedProcess()

        monkeypatch.setattr("subprocess.run", mock_subprocess_run)
        try:
            set_file_fn = self.get_tool_fn(server, "set_file")
            await set_file_fn(jsx_file_path)
            select_fn = self.get_tool_fn(server, "select")
            select_result = await select_fn(1, 7)
            overwrite_fn = self.get_tool_fn(server, "overwrite")
            invalid_jsx = {
                "lines": [
                    "import React from 'react';",
                    "",
                    "function BrokenComponent() {",
                    "  return <div>Missing closing tag<div>;",
                    "}",
                    "",
                    "export default BrokenComponent;",
                ]
            }
            result = await overwrite_fn(new_lines=invalid_jsx)
            assert "error" in result
            assert "JavaScript syntax error:" in result["error"]
            with open(jsx_file_path, "r") as f:
                file_content = f.read()
            assert file_content == valid_jsx_content
        finally:
            if os.path.exists(jsx_file_path):
                os.unlink(jsx_file_path)

    @pytest.mark.asyncio
    async def test_generate_diff_preview(self):
        """Test the generate_diff_preview function directly."""
        original_lines = ["Line 1", "Line 2", "Line 3", "Line 4", "Line 5"]
        modified_lines = [
            "Line 1",
            "Modified Line 2",
            "New Line",
            "Line 3",
            "Line 4",
            "Line 5",
        ]

        # Testing replacement in the middle of the file
        result = generate_diff_preview(original_lines, modified_lines, 2, 3)

        # Verify the result contains the expected diff_lines key
        assert "diff_lines" in result

        # Get and examine the content of diff_lines
        diff_lines_list = result["diff_lines"]

        # The diff_lines should be a list of tuples, let's check its structure
        # First verify we have the expected number of elements
        assert len(diff_lines_list) > 0

        # Check that we have context lines before the change
        # The first element should be the context line with line number 1
        assert any(item for item in diff_lines_list if item[0] == 1)

        # Check for removed lines with minus prefix
        assert any(item for item in diff_lines_list if item[0] == "-2")
        assert any(item for item in diff_lines_list if item[0] == "-3")

        # Check for added lines with plus prefix
        # There should be one entry containing the modified content
        added_lines = [
            item
            for item in diff_lines_list
            if isinstance(item[0], str) and item[0].startswith("+")
        ]
        assert len(added_lines) > 0

        # Verify context after the change (line 4 and 5)
        assert any(item for item in diff_lines_list if item[0] == 4)
        assert any(item for item in diff_lines_list if item[0] == 5)

    @pytest.fixture
    def python_test_file(self):
        """Create a Python test file with various functions and methods for testing find_function."""
        content = '''import os

def simple_function():
    """A simple function."""
    return "Hello, world!"

@decorator1
@decorator2
def decorated_function(a, b=None):
    """A function with decorators."""
    if b is None:
        b = a * 2
    return a + b

class TestClass:
    """A test class with methods."""
    
    def __init__(self, value):
        self.value = value
    
    def instance_method(self, x):
        """An instance method."""
        return self.value * x
    
    @classmethod
    def class_method(cls, y):
        """A class method."""
        return cls(y)
    
    @staticmethod
    def static_method(z):
        """A static method."""
        return z ** 2

def outer_function(param):
    """A function containing a nested function."""
    
    def inner_function(inner_param):
        """A nested function."""
        return inner_param + param
    
    return inner_function(param * 2)
'''
        with tempfile.NamedTemporaryFile(mode="w+", suffix=".py", delete=False) as f:
            f.write(content)
            temp_path = f.name
        yield temp_path
        if os.path.exists(temp_path):
            os.unlink(temp_path)

    @pytest.mark.asyncio
    async def test_find_function_no_file_set(self, server):
        """Test find_function when no file is set."""
        find_function_fn = self.get_tool_fn(server, "find_function")
        result = await find_function_fn(function_name="test")
        assert "error" in result
        assert "No file path is set" in result["error"]

    @pytest.mark.asyncio
    async def test_find_function_non_supported_file(self, server, temp_file):
        """Test find_function with a non-supported file type."""
        set_file_fn = self.get_tool_fn(server, "set_file")
        await set_file_fn(temp_file)
        find_function_fn = self.get_tool_fn(server, "find_function")
        result = await find_function_fn(function_name="test")
        assert "error" in result
        assert (
            "This tool only works with Python (.py) or JavaScript/JSX (.js, .jsx) files"
            in result["error"]
        )

    @pytest.mark.asyncio
    async def test_find_function_simple(self, server, python_test_file):
        """Test find_function with a simple function."""
        set_file_fn = self.get_tool_fn(server, "set_file")
        await set_file_fn(python_test_file)
        find_function_fn = self.get_tool_fn(server, "find_function")
        result = await find_function_fn(function_name="simple_function")
        assert "status" in result
        assert result["status"] == "success"
        assert "lines" in result
        assert "start_line" in result
        assert "end_line" in result

        # Check that the correct function is returned
        function_text = "".join(line[1] for line in result["lines"])
        assert "def simple_function():" in function_text
        assert "A simple function" in function_text
        assert 'return "Hello, world!"' in function_text

    @pytest.mark.asyncio
    async def test_find_function_decorated(self, server, python_test_file):
        """Test find_function with a decorated function."""
        set_file_fn = self.get_tool_fn(server, "set_file")
        await set_file_fn(python_test_file)
        find_function_fn = self.get_tool_fn(server, "find_function")
        result = await find_function_fn(function_name="decorated_function")
        assert result["status"] == "success"

        # Check that the decorators are included
        function_lines = [line[1] for line in result["lines"]]
        assert any("@decorator1" in line for line in function_lines)
        assert any("@decorator2" in line for line in function_lines)
        assert any(
            "def decorated_function(a, b=None):" in line for line in function_lines
        )

    @pytest.mark.asyncio
    async def test_find_function_method(self, server, python_test_file):
        """Test find_function with a class method."""
        set_file_fn = self.get_tool_fn(server, "set_file")
        await set_file_fn(python_test_file)
        find_function_fn = self.get_tool_fn(server, "find_function")
        result = await find_function_fn(function_name="instance_method")
        assert result["status"] == "success"

        # Check that the method is correctly identified
        function_text = "".join(line[1] for line in result["lines"])
        assert "def instance_method(self, x):" in function_text
        assert "An instance method" in function_text
        assert "return self.value * x" in function_text

    @pytest.mark.asyncio
    async def test_find_function_static_method(self, server, python_test_file):
        """Test find_function with a static method."""
        set_file_fn = self.get_tool_fn(server, "set_file")
        await set_file_fn(python_test_file)
        find_function_fn = self.get_tool_fn(server, "find_function")
        result = await find_function_fn(function_name="static_method")
        assert result["status"] == "success"

        # Check that the decorator and method are included
        function_lines = [line[1] for line in result["lines"]]
        assert any("@staticmethod" in line for line in function_lines)
        assert any("def static_method(z):" in line for line in function_lines)

    @pytest.mark.asyncio
    async def test_find_function_not_found(self, server, python_test_file):
        """Test find_function with a non-existent function."""
        set_file_fn = self.get_tool_fn(server, "set_file")
        await set_file_fn(python_test_file)
        find_function_fn = self.get_tool_fn(server, "find_function")
        result = await find_function_fn(function_name="nonexistent_function")
        assert "error" in result
        assert "not found in the file" in result["error"]

    @pytest.mark.asyncio
    async def test_protect_paths_env_variable(self, monkeypatch):
        """Test that the PROTECTED_PATHS environment variable is correctly processed."""
        # Set up the environment variable with test paths
        monkeypatch.setenv(
            "PROTECTED_PATHS",
            "*.secret,.env*,config*.json,*sensitive*,/etc/shadow,/home/user/.ssh/id_rsa",
        )

        # Create a new server instance which should read the environment variable
        server = TextEditorServer()

        # Verify the protected_paths list is populated correctly
        assert len(server.protected_paths) == 6
        assert "*.secret" in server.protected_paths
        assert ".env*" in server.protected_paths
        assert "config*.json" in server.protected_paths
        assert "*sensitive*" in server.protected_paths
        assert "/etc/shadow" in server.protected_paths
        assert "/home/user/.ssh/id_rsa" in server.protected_paths

    @pytest.mark.asyncio
    async def test_protect_paths_empty_env_variable(self, monkeypatch):
        """Test that an empty PROTECTED_PATHS environment variable is handled correctly."""
        # Set up an empty environment variable
        monkeypatch.setenv("PROTECTED_PATHS", "")

        # Create a new server instance
        server = TextEditorServer()

        # Verify the protected_paths list is empty
        assert len(server.protected_paths) == 0

    @pytest.mark.asyncio
    async def test_protect_paths_trimming(self, monkeypatch):
        """Test that whitespace in PROTECTED_PATHS items is properly trimmed."""
        # Set up the environment variable with whitespace
        monkeypatch.setenv(
            "PROTECTED_PATHS", " *.secret , /etc/shadow ,  /home/user/.ssh/id_rsa "
        )

        # Create a new server instance
        server = TextEditorServer()

        # Get set_file tool for testing
        set_file_fn = self.get_tool_fn(server, "set_file")

        # Mock os.path.isfile to return True for our test path
        def mock_isfile(path):
            return True

        monkeypatch.setattr(os.path, "isfile", mock_isfile)

        # Test access denied for a path matching a trimmed pattern
        result = await set_file_fn("/home/user/.ssh/id_rsa")
        assert "Error: Access to '/home/user/.ssh/id_rsa' is denied" in result

    @pytest.mark.asyncio
    async def test_find_function_nested(self, server, python_test_file):
        """Test find_function with nested functions."""
        set_file_fn = self.get_tool_fn(server, "set_file")
        await set_file_fn(python_test_file)
        find_function_fn = self.get_tool_fn(server, "find_function")

        # Test finding the outer function
        result = await find_function_fn(function_name="outer_function")
        assert result["status"] == "success"
        function_text = "".join(line[1] for line in result["lines"])
        assert "def outer_function(param):" in function_text
        assert "def inner_function(inner_param):" in function_text

        # Test finding the inner function (this may or may not work depending on implementation)
        # AST might not directly support finding nested functions
        # This test is designed to document current behavior, not necessarily assert correctness
        inner_result = await find_function_fn(function_name="inner_function")
        # If it finds the inner function, check it's correct
        if "status" in inner_result and inner_result["status"] == "success":
            inner_text = "".join(line[1] for line in inner_result["lines"])
            assert "def inner_function(inner_param):" in inner_text
        # Otherwise, it should return an error that the function wasn't found
        else:
            assert "error" in inner_result
            assert "not found in the file" in inner_result["error"]

    @pytest.mark.asyncio
    async def test_find_function_parsing_error(self, server):
        """Test find_function with a file that can't be parsed due to syntax errors."""
        with tempfile.NamedTemporaryFile(mode="w+", suffix=".py", delete=False) as f:
            f.write(
                "def broken_function(  # Syntax error: missing parenthesis\n    pass\n"
            )
            invalid_py_path = f.name
        try:
            set_file_fn = self.get_tool_fn(server, "set_file")
            await set_file_fn(invalid_py_path)
            find_function_fn = self.get_tool_fn(server, "find_function")
            result = await find_function_fn(function_name="broken_function")
            assert "error" in result
            assert "Error finding function" in result["error"]
        finally:
            if os.path.exists(invalid_py_path):
                os.unlink(invalid_py_path)

    @pytest.mark.asyncio
    async def test_find_function_javascript(
        self, server, javascript_test_file, monkeypatch
    ):
        """Test find_function with JavaScript functions."""

        # Mock subprocess.run to avoid external dependency in tests
        def mock_subprocess_run(*args, **kwargs):
            class MockCompletedProcess:
                def __init__(self):
                    self.returncode = 0
                    self.stderr = ""
                    self.stdout = ""

            return MockCompletedProcess()

        monkeypatch.setattr("subprocess.run", mock_subprocess_run)

        set_file_fn = self.get_tool_fn(server, "set_file")
        await set_file_fn(javascript_test_file)
        find_function_fn = self.get_tool_fn(server, "find_function")

        # Test regular function
        result = await find_function_fn(function_name="simpleFunction")
        assert result["status"] == "success"
        # Get all the lines of the function
        function_lines = [line[1] for line in result["lines"]]
        assert any("function simpleFunction()" in line for line in function_lines)
        assert any("console.log('Hello world')" in line for line in function_lines)

        # Test arrow function
        result = await find_function_fn(function_name="arrowFunction")
        assert result["status"] == "success"
        function_lines = [line[1] for line in result["lines"]]
        assert any("const arrowFunction = (a, b) =>" in line for line in function_lines)

        # Test async function
        result = await find_function_fn(function_name="asyncFunction")
        assert result["status"] == "success"
        function_lines = [line[1] for line in result["lines"]]
        assert any("async function asyncFunction()" in line for line in function_lines)

        # Test hook-style function
        result = await find_function_fn(function_name="useCustomHook")
        assert result["status"] == "success"
        function_lines = [line[1] for line in result["lines"]]
        assert any(
            "const useCustomHook = useCallback" in line for line in function_lines
        )

        result = await find_function_fn(function_name="methodFunction")

        function_lines = [line[1] for line in result["lines"]]
        assert any("methodFunction(x, y)" in line for line in function_lines)

        result = await find_function_fn(function_name="nonExistentFunction")
        assert "error" in result
        assert "not found in the file" in result["error"]

    @pytest.mark.asyncio
    async def test_find_function_jsx(self, server, jsx_test_file, monkeypatch):
        """Test find_function with JSX/React component functions."""

        def mock_subprocess_run(*args, **kwargs):
            class MockCompletedProcess:
                def __init__(self):
                    self.returncode = 0
                    self.stderr = ""
                    self.stdout = ""

            return MockCompletedProcess()

        monkeypatch.setattr("subprocess.run", mock_subprocess_run)

        set_file_fn = self.get_tool_fn(server, "set_file")
        await set_file_fn(jsx_test_file)
        find_function_fn = self.get_tool_fn(server, "find_function")

        # Test regular function component
        result = await find_function_fn(function_name="SimpleComponent")
        assert result["status"] == "success"
        # Get all the lines of the component
        function_lines = [line[1] for line in result["lines"]]
        # Just check that the function name is found
        assert "SimpleComponent" in "".join(function_lines)

        # Test arrow function component
        result = await find_function_fn(function_name="ArrowComponent")
        assert result["status"] == "success"
        # Get all the lines of the component
        function_lines = [line[1] for line in result["lines"]]
        # Only check that we found the function declaration
        assert "ArrowComponent" in "".join(function_lines)

        # Test component with nested function
        result = await find_function_fn(function_name="ParentComponent")
        assert result["status"] == "success"
        # Get all the lines of the component
        function_lines = [line[1] for line in result["lines"]]
        assert any("function ParentComponent()" in line for line in function_lines)
        assert any("function handleClick()" in line for line in function_lines)

        # Test higher order component
        result = await find_function_fn(function_name="withLogger")
        assert result["status"] == "success"
        # Get all the lines of the component
        function_lines = [line[1] for line in result["lines"]]
        assert any("function withLogger(Component)" in line for line in function_lines)
        assert any(
            "return function EnhancedComponent(props)" in line
            for line in function_lines
        )

        # Test nested function may or may not work depending on implementation
        result = await find_function_fn(function_name="handleClick")
        if "status" in result and result["status"] == "success":
            function_lines = [line[1] for line in result["lines"]]
            assert any("function handleClick()" in line for line in function_lines)
        else:
            assert "error" in result

        # Test non-existent function
        result = await find_function_fn(function_name="nonExistentComponent")
        assert "error" in result
        assert "not found in the file" in result["error"]

    @pytest.mark.asyncio
    async def test_find_function_js_with_disabled_check(self, server, monkeypatch):
        """Test find_function with disabled JavaScript syntax checking."""
        # Create a server with disabled JS syntax checking
        monkeypatch.setenv("ENABLE_JS_SYNTAX_CHECK", "0")
        server_no_js_check = TextEditorServer()

        # Create a basic JavaScript file
        js_content = "function testFunc() { return 'test'; }"
        with tempfile.NamedTemporaryFile(mode="w+", suffix=".js", delete=False) as f:
            f.write(js_content)
            js_file_path = f.name

        try:
            # This test ensures find_function still works even when JavaScript
            # syntax checking is disabled for overwrite operations
            set_file_fn = self.get_tool_fn(server_no_js_check, "set_file")
            await set_file_fn(js_file_path)
            find_function_fn = self.get_tool_fn(server_no_js_check, "find_function")

            result = await find_function_fn(function_name="testFunc")
            assert result["status"] == "success"
            function_lines = [line[1] for line in result["lines"]]
            assert any("function testFunc()" in line for line in function_lines)
        finally:
            if os.path.exists(js_file_path):
                os.unlink(js_file_path)

    @pytest.mark.asyncio
    async def test_listdir_tool(self, server, temp_file, monkeypatch):
        """Test the listdir tool functionality."""
        # Create a test directory
        with tempfile.TemporaryDirectory() as temp_dir:
            # Create some test files in the directory
            test_files = ["file1.txt", "file2.py", "file3.js"]
            for file_name in test_files:
                file_path = os.path.join(temp_dir, file_name)
                with open(file_path, "w") as f:
                    f.write(f"Content for {file_name}")

            # Test the listdir tool
            listdir_fn = self.get_tool_fn(server, "listdir")
            result = await listdir_fn(dirpath=temp_dir)

            # Check that the result contains the expected data
            assert "filenames" in result
            assert "path" in result
            assert result["path"] == temp_dir

            # Check that all test files are in the result
            for file_name in test_files:
                assert file_name in result["filenames"]

    @pytest.mark.asyncio
    async def test_listdir_error_not_directory(self, server, temp_file):
        """Test the listdir tool with a path that is not a directory."""
        # Use the temp_file fixture which is a file, not a directory
        listdir_fn = self.get_tool_fn(server, "listdir")
        result = await listdir_fn(dirpath=temp_file)

        # Check that an appropriate error is returned
        assert "error" in result
        assert "not a directory" in result["error"].lower()
        assert "path" in result
        assert result["path"] == temp_file

    @pytest.mark.asyncio
    async def test_listdir_error_nonexistent_path(self, server):
        """Test the listdir tool with a non-existent path."""
        # Create a path that doesn't exist
        nonexistent_path = "/path/that/does/not/exist"

        # Test the listdir tool with the non-existent path
        listdir_fn = self.get_tool_fn(server, "listdir")
        result = await listdir_fn(dirpath=nonexistent_path)

        # Check that an appropriate error is returned
        assert "error" in result
        assert "unexpected error" in result["error"].lower()

    @pytest.mark.asyncio
    async def test_duckdb_usage_stats_enabled(self, monkeypatch):
        """Test that usage stats are enabled when environment variable is set."""
        # Mock environment variables to enable stats
        monkeypatch.setenv("DUCKDB_USAGE_STATS", "1")
        monkeypatch.setenv("STATS_DB_PATH", ":memory:")

        # Create a server with usage stats enabled
        server = TextEditorServer()

        # Check that usage stats are enabled
        assert server.usage_stats_enabled is True
        assert server.stats_db_path == ":memory:"

        # Verify the decorator wrapper was applied
        assert server.mcp.tool != FastMCP.tool

    @pytest.mark.asyncio
    async def test_duckdb_usage_stats_disabled(self, monkeypatch):
        """Test that usage stats are disabled by default."""
        # Mock environment variables to explicitly disable stats
        monkeypatch.setenv("DUCKDB_USAGE_STATS", "0")

        # Create a server with usage stats disabled
        server = TextEditorServer()

        # Check that usage stats are disabled
        assert server.usage_stats_enabled is False
        assert not hasattr(server, "stats_db_path")

    @pytest.mark.asyncio
    async def test_find_js_function_babel(
        self, server, javascript_test_file, monkeypatch
    ):
        """Test the _find_js_function_babel method for JavaScript parsing."""

        # Create a mock subprocess result with Babel output
        class MockCompletedProcess:
            def __init__(self):
                self.returncode = 0
                # Mock Babel output with function location data
                self.stdout = """FUNCTION_LOCATIONS: {
                    "simpleFunction": {
                        "start": {"line": 5, "column": 0},
                        "end": {"line": 8, "column": 1}
                    }
                }"""
                self.stderr = ""

        # Mock subprocess.run to return our mock data
        monkeypatch.setattr(
            "subprocess.run", lambda *args, **kwargs: MockCompletedProcess()
        )

        # Set up the server with the test file
        set_file_fn = self.get_tool_fn(server, "set_file")
        await set_file_fn(javascript_test_file)

        # Call _find_js_function_babel directly
        result = server._find_js_function_babel(
            "simpleFunction",
            "function simpleFunction() {}\n",
            ["function simpleFunction() {}\n"],
        )

        # Verify the result
        assert result is not None
        assert "status" in result
        assert result["status"] == "success"
        assert "parser" in result
        assert result["parser"] == "babel"
        assert "start_line" in result
        assert result["start_line"] == 5
        assert "end_line" in result
        assert result["end_line"] == 8

    @pytest.mark.asyncio
    async def test_find_js_function_babel_no_match(
        self, server, javascript_test_file, monkeypatch
    ):
        """Test _find_js_function_babel when no matching function is found."""

        # Create a mock subprocess result with Babel output for a different function
        class MockCompletedProcess:
            def __init__(self):
                self.returncode = 0
                # Mock Babel output with function location data
                self.stdout = """FUNCTION_LOCATIONS: {
                    "otherFunction": {
                        "start": {"line": 10, "column": 0},
                        "end": {"line": 12, "column": 1}
                    }
                }"""
                self.stderr = ""

        # Mock subprocess.run to return our mock data
        monkeypatch.setattr(
            "subprocess.run", lambda *args, **kwargs: MockCompletedProcess()
        )

        # Set up the server with the test file
        set_file_fn = self.get_tool_fn(server, "set_file")
        await set_file_fn(javascript_test_file)

        # Call _find_js_function_babel for a function that doesn't exist in the mock output
        result = server._find_js_function_babel(
            "simpleFunction",
            "function simpleFunction() {}\n",
            ["function simpleFunction() {}\n"],
        )

        # Should return None when the function is not found
        assert result is None

    @pytest.fixture
    def javascript_test_file(self):
        """Create a JavaScript test file with various functions for testing find_function."""
        content = """// Sample JavaScript file with different function types

// Regular function declaration
function simpleFunction() {
  console.log('Hello world');
  return 42;
}

// Arrow function expression
const arrowFunction = (a, b) => {
  const sum = a + b;
  return sum;
};

// Object with method
const obj = {
  methodFunction(x, y) {
    return x * y;
  },

  // Object method as arrow function
  arrowMethod: (z) => {
    return z * z;
  }
};

// Async function
async function asyncFunction() {
  return await Promise.resolve('done');
}

// React hook style function
const useCustomHook = useCallback((value) => {
  return value.toUpperCase();
}, []);

// Class with methods
class TestClass {
  constructor(value) {
    this.value = value;
  }

  instanceMethod() {
    return this.value;
  }

  static staticMethod() {
    return 'static';
  }
}
"""
        with tempfile.NamedTemporaryFile(mode="w+", suffix=".js", delete=False) as f:
            f.write(content)
            temp_path = f.name
        yield temp_path
        if os.path.exists(temp_path):
            os.unlink(temp_path)

    @pytest.fixture
    def jsx_test_file(self):
        """Create a JSX test file with various component functions for testing find_function."""
        content = """import React, { useState, useEffect } from 'react';

// Function component
function SimpleComponent() {
  return <div>Hello World</div>;
}

// Arrow function component with props
const ArrowComponent = ({ name }) => {
  const [count, setCount] = useState(0);

  useEffect(() => {
    document.title = `${name}: ${count}`;
  }, [name, count]);

  return (
    <div>
      <h1>Hello {name}</h1>
      <button onClick={() => setCount(count + 1)}>
        Count: {count}
      </button>
    </div>
  );
};

// Component with nested function
function ParentComponent() {
  function handleClick() {
    console.log('Button clicked');
  }

  return <button onClick={handleClick}>Click me</button>;
}

// Higher order component
function withLogger(Component) {
  return function EnhancedComponent(props) {
    console.log('Component rendered with props:', props);
    return <Component {...props} />;
  };
}

export default SimpleComponent;
"""
        with tempfile.NamedTemporaryFile(mode="w+", suffix=".jsx", delete=False) as f:
            f.write(content)
            temp_path = f.name
        yield temp_path
        if os.path.exists(temp_path):
            os.unlink(temp_path)

    @pytest.mark.asyncio
    async def test_run_tests_basic(self, server, monkeypatch):
        """Test the basic functionality of run_tests."""

        # Mock subprocess.run to simulate pytest execution
        def mock_subprocess_run(*args, **kwargs):
            class MockCompletedProcess:
                def __init__(self):
                    self.returncode = 0
                    self.stdout = "===== 5 passed in 0.12s ====="
                    self.stderr = ""

            return MockCompletedProcess()

        monkeypatch.setattr("subprocess.run", mock_subprocess_run)
        monkeypatch.setattr("datetime.datetime", MockDateTime)

        # Run the test function
        run_tests_fn = self.get_tool_fn(server, "run_tests")
        result = await run_tests_fn()
        print("Result:", result)

        # Verify the result
        assert result["status"] == "success"
        assert result["returncode"] == 0
        assert "5 passed" in result["stdout"]
        assert result["duration"] == 0.5  # From our mock
        assert "python -m pytest" in result["command"]

    @pytest.mark.asyncio
    async def test_run_tests_with_path(self, server, monkeypatch):
        """Test run_tests with a specific test path."""
        expected_cmd = ""

        def mock_subprocess_run(*args, **kwargs):
            nonlocal expected_cmd
            expected_cmd = " ".join(args[0])

            class MockCompletedProcess:
                def __init__(self):
                    self.returncode = 0
                    self.stdout = "===== 3 passed in 0.05s ====="
                    self.stderr = ""

            return MockCompletedProcess()

        monkeypatch.setattr("subprocess.run", mock_subprocess_run)
        monkeypatch.setattr("datetime.datetime", MockDateTime)

        test_path = "tests/test_module.py"
        run_tests_fn = self.get_tool_fn(server, "run_tests")
        result = await run_tests_fn(test_path=test_path)

        assert result["status"] == "success"
        assert result["returncode"] == 0
        assert test_path in expected_cmd
        assert "3 passed" in result["stdout"]

    @pytest.mark.asyncio
    async def test_run_tests_with_test_name(self, server, monkeypatch):
        """Test run_tests with a specific test name."""
        expected_cmd = ""

        def mock_subprocess_run(*args, **kwargs):
            nonlocal expected_cmd
            expected_cmd = " ".join(args[0])

            class MockCompletedProcess:
                def __init__(self):
                    self.returncode = 0
                    self.stdout = "===== 1 passed in 0.01s ====="
                    self.stderr = ""

            return MockCompletedProcess()

        monkeypatch.setattr("subprocess.run", mock_subprocess_run)
        monkeypatch.setattr("datetime.datetime", MockDateTime)

        test_name = "test_specific_function"
        run_tests_fn = self.get_tool_fn(server, "run_tests")
        result = await run_tests_fn(test_name=test_name)

        assert result["status"] == "success"
        assert result["returncode"] == 0
        assert f"-k {test_name}" in expected_cmd
        assert "1 passed" in result["stdout"]

    @pytest.mark.asyncio
    async def test_run_tests_verbose(self, server, monkeypatch):
        """Test run_tests with verbose flag."""
        expected_cmd = ""

        def mock_subprocess_run(*args, **kwargs):
            nonlocal expected_cmd
            expected_cmd = " ".join(args[0])

            class MockCompletedProcess:
                def __init__(self):
                    self.returncode = 0
                    self.stdout = "===== verbose output ====="
                    self.stderr = ""

            return MockCompletedProcess()

        monkeypatch.setattr("subprocess.run", mock_subprocess_run)
        monkeypatch.setattr("datetime.datetime", MockDateTime)

        run_tests_fn = self.get_tool_fn(server, "run_tests")
        result = await run_tests_fn(verbose=True)

        assert result["status"] == "success"
        assert result["returncode"] == 0
        assert "-v" in expected_cmd

    @pytest.mark.asyncio
    async def test_run_tests_collect_only(self, server, monkeypatch):
        """Test run_tests with collect_only flag."""
        expected_cmd = ""

        def mock_subprocess_run(*args, **kwargs):
            nonlocal expected_cmd
            expected_cmd = " ".join(args[0])

            class MockCompletedProcess:
                def __init__(self):
                    self.returncode = 0
                    self.stdout = "collected 10 items"
                    self.stderr = ""

            return MockCompletedProcess()

        monkeypatch.setattr("subprocess.run", mock_subprocess_run)
        monkeypatch.setattr("datetime.datetime", MockDateTime)

        run_tests_fn = self.get_tool_fn(server, "run_tests")
        result = await run_tests_fn(collect_only=True)

        assert result["status"] == "success"
        assert result["returncode"] == 0
        assert "--collect-only" in expected_cmd
        assert "collected 10 items" in result["stdout"]

    @pytest.mark.asyncio
    async def test_run_tests_failure(self, server, monkeypatch):
        """Test run_tests when tests fail."""

        def mock_subprocess_run(*args, **kwargs):
            class MockCompletedProcess:
                def __init__(self):
                    self.returncode = 1
                    self.stdout = "===== 2 failed, 3 passed in 0.05s ====="
                    self.stderr = "E       AssertionError: expected 5 but got 6"

            return MockCompletedProcess()

        monkeypatch.setattr("subprocess.run", mock_subprocess_run)
        monkeypatch.setattr("datetime.datetime", MockDateTime)

        run_tests_fn = self.get_tool_fn(server, "run_tests")
        result = await run_tests_fn()

        assert result["status"] == "failure"
        assert result["returncode"] == 1
        assert "2 failed" in result["stdout"]
        assert "AssertionError" in result["stderr"]

    @pytest.mark.asyncio
    async def test_run_tests_error(self, server, monkeypatch):
        """Test run_tests when an exception occurs."""

        def mock_subprocess_run(*args, **kwargs):
            raise Exception("Command execution failed")

        monkeypatch.setattr("subprocess.run", mock_subprocess_run)
        run_tests_fn = self.get_tool_fn(server, "run_tests")
        result = await run_tests_fn()

        assert result["status"] == "error"
        assert "error" in result
        assert "Command execution failed" in result["error"]

    @pytest.mark.asyncio
    async def test_run_tests_with_env_python_venv(self, server, monkeypatch):
        """Test run_tests using Python virtual environment from environment variable."""
        expected_cmd = ""

        def mock_subprocess_run(*args, **kwargs):
            nonlocal expected_cmd
            expected_cmd = " ".join(args[0])

            class MockCompletedProcess:
                def __init__(self):
                    self.returncode = 0
                    self.stdout = "===== All tests passed ====="
                    self.stderr = ""

            return MockCompletedProcess()

        monkeypatch.setattr("subprocess.run", mock_subprocess_run)
        monkeypatch.setattr("datetime.datetime", MockDateTime)

        # Set the python_venv at the server level (environment variable simulation)
        env_venv_path = "/env/path/to/python"
        server.python_venv = env_venv_path

        run_tests_fn = self.get_tool_fn(server, "run_tests")
        result = await run_tests_fn()

        assert result["status"] == "success"
        assert result["returncode"] == 0
        assert env_venv_path in expected_cmd
        assert env_venv_path == expected_cmd.split()[0]  # Should be first in command


# Helper class to mock datetime for consistent duration in tests
class MockDateTime:
    _now_counter = 0

    @classmethod
    def now(cls):
        cls._now_counter += 1
        return cls.MockDatetimeValue(0 if cls._now_counter % 2 == 1 else 0.5)

    class MockDatetimeValue:
        def __init__(self, value):
            self.value = value

        def __sub__(self, other):
            return MockDateTime.MockTimeDelta(0.5)  # Always return 0.5 seconds

    class MockTimeDelta:
        def __init__(self, seconds):
            self.seconds = seconds

        def total_seconds(self):
            return self.seconds



"""
MCP Text Editor Package

A Python-based text editor server built with FastMCP that provides tools 
for file operations. This server enables reading, editing, and managing 
text files through a standardized API.
"""

__version__ = "0.1.0"



import hashlib
import logging
import os
import re
import subprocess
import tempfile
import ast
import tokenize
import fnmatch
import io
import datetime
import json
import inspect
import functools
from typing import Optional, Dict, Any, Union, Literal
import argparse
try:
    import black
    from black.report import NothingChanged
    BLACK_AVAILABLE = True
except ImportError:
    BLACK_AVAILABLE = False
    NothingChanged = Exception  # Fallback for when black is not available
from fastmcp import FastMCP
import duckdb


logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)

logger = logging.getLogger("text_editor")


def calculate_id(text: str, start: int = None, end: int = None) -> str:
    """
    Calculate a unique ID for content verification based on the text content.

    The ID is formed by combining a line prefix (if line numbers are provided)
    with a truncated SHA-256 id of the content. This allows quick verification
    that content hasn't changed between operations.

    Args:
        text (str): Content to generate ID for
        start (Optional[int]): Starting line number for the content
        end (Optional[int]): Ending line number for the content
    Returns:
        str: ID string in format: [LinePrefix]-[Truncatedid]
             Example: "L10-15-a7" for content spanning lines 10-15
             Example: "L5-b3" for content on line 5 only
    """
    prefix = ""
    if start and end:
        prefix = f"L{start}-{end}-"
        if start == end:
            prefix = f"L{start}-"

    return f"{prefix}{hashlib.sha256(text.encode()).hexdigest()[:2]}"


def generate_diff_preview(
    original_lines: list, modified_lines: list, start: int, end: int
) -> dict:
    """
    Generate a diff preview comparing original and modified content.

    Args:
        original_lines (list): List of original file lines
        modified_lines (list): List of modified file lines
        start (int): Start line number of the edit (1-based)
        end (int): End line number of the edit (1-based)

    Returns:
        dict: A dictionary with keys prefixed with + or - to indicate additions/deletions
              Format: [("-1", "removed line"), ("+1", "added line")]
    """
    diffs = []
    # Add some context lines before the change
    context_start = max(0, start - 1 - 3)  # 3 lines of context before
    for i in range(context_start, start - 1):
        diffs.append((i + 1, original_lines[i].rstrip()))
    # Show removed lines
    for i in range(start - 1, end):
        diffs.append((f"-{i+1}", original_lines[i].rstrip()))

    # Show added lines
    new_content = "".join(
        modified_lines[
            start - 1 : start
            - 1
            + len(modified_lines)
            - len(original_lines)
            + (end - (start - 1))
        ]
    )
    new_lines = new_content.splitlines()
    for i, line in enumerate(new_lines):
        diffs.append((f"+{start+i}", line))
    context_end = min(len(original_lines), end + 3)  # 3 lines of context after
    for i in range(end, context_end):
        diffs.append((i + 1, original_lines[i].rstrip()))
    return {
        "diff_lines": diffs,
    }


def create_logging_tool_decorator(original_decorator, log_callback):
    """
    Create a wrapper around the FastMCP tool decorator that logs tool usage.

    Args:
        original_decorator: The original FastMCP tool decorator
        log_callback: A callback function that will be called with (tool_name, args_dict, response)

    Returns:
        A wrapped decorator function that logs tool usage
    """

    def tool_decorator_with_logging(*args, **kwargs):
        # Get the original decorator with args
        wrapped_decorator = original_decorator(*args, **kwargs)

        def wrapper(func):
            # Create our logging wrapper
            @functools.wraps(func)  # Preserve func's metadata
            async def logged_func(*func_args, **func_kwargs):
                tool_name = func.__name__
                # Convert args to dict for logging
                args_dict = {}
                if func_args and len(func_args) > 0:
                    # Get parameter names from function signature
                    sig = inspect.signature(func)
                    param_names = list(sig.parameters.keys())
                    # Skip self parameter if it exists
                    if param_names and param_names[0] == "self":
                        param_names = param_names[1:]
                    for i, arg in enumerate(func_args):
                        if i < len(param_names):
                            args_dict[param_names[i]] = (
                                str(arg) if isinstance(arg, (bytes, bytearray)) else arg
                            )
                args_dict.update(func_kwargs)

                # Call the original function
                response = await func(*func_args, **func_kwargs)

                # Log the tool usage with the response
                log_callback(tool_name, args_dict, response)

                # Return the response
                return response

            # Apply the original decorator to our logged function
            wrapped_func = wrapped_decorator(logged_func)
            return wrapped_func

        return wrapper

    return tool_decorator_with_logging


class TextEditorServer:
    """
    A server implementation for a text editor application using FastMCP.

    This class provides a comprehensive set of tools for manipulating text files in a
    controlled and safe manner. It implements a structured editing workflow with content
    verification to prevent conflicts during editing operations.

    Primary features:
    - File management: Setting active file, creating new files, and deleting files
    - Content access: Reading full file content or specific line ranges
    - Text search: Finding lines matching specific text patterns
    - Safe editing: Two-step edit process with diff preview and confirmation
    - Syntax validation: Automatic syntax checking for Python and JavaScript files
    - Protected files: Prevent access to sensitive files via pattern matching

    The server uses content hashing to generate unique IDs that ensure file content
    integrity during editing operations. All tools are registered with FastMCP for
    remote procedure calling.

    Edit workflow:
    1. Select content range with the select() tool to identify lines for editing
    2. Propose changes with overwrite() to generate a diff preview
    3. Use confirm() to apply or cancel() to discard the pending modifications

    Attributes:
        mcp (FastMCP): The MCP server instance for handling tool registrations
        max_select_lines (int): Maximum number of lines that can be edited with ID verification
        enable_js_syntax_check (bool): Whether JavaScript syntax checking is enabled
        protected_paths (list): List of file patterns and paths that are restricted from access
        current_file_path (str, optional): Path to the currently active file
        selected_start (int, optional): Start line of the current selection
        selected_end (int, optional): End line of the current selection
        selected_id (str, optional): ID of the current selection for verification
        pending_modified_lines (list, optional): Pending modified lines for preview before committing
        pending_diff (dict, optional): Diff preview of pending changes
    """

    def __init__(self):
        # Initialize MCP server
        self.mcp = FastMCP("text-editor")

        # Initialize DuckDB for usage statistics if enabled
        self.usage_stats_enabled = os.getenv("DUCKDB_USAGE_STATS", "0").lower() in [
            "1",
            "true",
            "yes",
        ]
        if self.usage_stats_enabled:
            self.stats_db_path = os.getenv("STATS_DB_PATH", "text_editor_stats.duckdb")
            self._init_stats_db()
            # Replace the original tool decorator with our wrapper
            # Making sure we preserve all the original functionality
            original_tool = self.mcp.tool
            logger.debug(
                {
                    "message": f"Setting up logging tool decorator using {self.stats_db_path}"
                }
            )
            self.mcp.tool = create_logging_tool_decorator(
                original_tool, self._log_tool_usage
            )
            logger.debug({"msg": "Logging tool decorator set up complete"})
        self.max_select_lines = int(os.getenv("MAX_SELECT_LINES", "50"))
        self.enable_js_syntax_check = os.getenv(
            "ENABLE_JS_SYNTAX_CHECK", "1"
        ).lower() in ["1", "true", "yes"]
        self.fail_on_python_syntax_error = os.getenv(
            "FAIL_ON_PYTHON_SYNTAX_ERROR", "1"
        ).lower() in ["1", "true", "yes"]
        self.fail_on_js_syntax_error = os.getenv(
            "FAIL_ON_JS_SYNTAX_ERROR", "0"
        ).lower() in ["1", "true", "yes"]
        self.protected_paths = (
            os.getenv("PROTECTED_PATHS", "").split(",")
            if os.getenv("PROTECTED_PATHS")
            else []
        )
        self.current_file_path = None
        self.selected_start = None
        self.selected_end = None
        self.selected_id = None
        self.pending_modified_lines = None
        self.pending_diff = None
        self.python_venv = os.getenv("PYTHON_VENV")

        self.register_tools()

    def _init_stats_db(self):
        """Initialize the DuckDB database for storing tool usage statistics."""
        logger.debug({"msg": f"Initializing stats database at {self.stats_db_path}"})
        try:
            # Connect to DuckDB and create the table if it doesn't exist
            with duckdb.connect(self.stats_db_path) as conn:
                logger.debug({"msg": "Connected to DuckDB for initialization"})
                conn.execute("""
                    CREATE TABLE IF NOT EXISTS tool_usage (
                        tool_name VARCHAR,
                        args JSON,
                        response JSON,
                        timestamp TIMESTAMP,
                        current_file VARCHAR,
                        request_id VARCHAR,
                        client_id VARCHAR
                    )
                """)
                conn.commit()
                logger.debug({"msg": "Database tables initialized successfully"})
        except Exception as e:
            logger.debug({"msg": f"Error initializing stats database: {str(e)}"})

    def _log_tool_usage(self, tool_name: str, args: dict, response=None):
        """
        Log tool usage to the DuckDB database if stats are enabled.
        This function is called by the decorator wrapper for each tool invocation.

        Args:
            tool_name (str): Name of the tool being used
            args (dict): Arguments passed to the tool
            response (dict, optional): Response returned by the tool
        """
        # Skip if stats are disabled
        if not hasattr(self, "usage_stats_enabled") or not self.usage_stats_enabled:
            return

        logger.debug({"message": f"Logging tool usage: {tool_name}"})
        client_id = None
        try:
            # Get request ID if available
            request_id = None
            if hasattr(self.mcp, "_mcp_server") and hasattr(
                self.mcp._mcp_server, "request_context"
            ):
                request_id = getattr(
                    self.mcp._mcp_server.request_context, "request_id", None
                )

                if hasattr(self.mcp, "_mcp_server") and hasattr(
                    self.mcp._mcp_server, "request_context"
                ):
                    # Access client_id from meta if available
                    if (
                        hasattr(self.mcp._mcp_server.request_context, "meta")
                        and self.mcp._mcp_server.request_context.meta
                    ):
                        client_id = getattr(
                            self.mcp._mcp_server.request_context.meta, "client_id", None
                        )

            # Safely convert args to serializable format
            safe_args = {}
            for key, value in args.items():
                # Skip large objects and convert non-serializable objects to strings
                if (
                    hasattr(value, "__len__")
                    and not isinstance(value, (str, dict, list, tuple))
                    and len(value) > 1000
                ):
                    safe_args[key] = f"<{type(value).__name__} of length {len(value)}>"
                else:
                    try:
                        # Test if value is JSON serializable
                        json.dumps({key: value})
                        safe_args[key] = value
                    except (TypeError, OverflowError):
                        # If not serializable, convert to string representation
                        safe_args[key] = f"<{type(value).__name__}>"

            # Format arguments as JSON
            args_json = json.dumps(safe_args)

            # Process response - since we're using JSON RPC, responses should already be serializable
            response_json = None
            if response is not None:
                try:
                    response_json = json.dumps(response)
                except (TypeError, OverflowError):
                    # Handle edge cases where something non-serializable might have been returned
                    logger.debug(
                        {
                            "message": f"Non-serializable response received from {tool_name}, converting to string representation"
                        }
                    )
                    if isinstance(response, dict):
                        # For dictionaries, process each value separately
                        safe_response = {}
                        for key, value in response.items():
                            try:
                                json.dumps({key: value})
                                safe_response[key] = value
                            except (TypeError, OverflowError):
                                safe_response[key] = f"<{type(value).__name__}>"
                        response_json = json.dumps(safe_response)
                    else:
                        # For non-dict types, store a basic representation
                        response_json = json.dumps({"result": str(response)})

            logger.debug(
                {"msg": f"Attempting to connect to DuckDB at {self.stats_db_path}"}
            )
            try:
                with duckdb.connect(self.stats_db_path) as conn:
                    logger.debug({"msg": f"Connected to DuckDB successfully"})
                    conn.execute(
                        """
                        INSERT INTO tool_usage (tool_name, args, response, timestamp, current_file, request_id, client_id)
                        VALUES (?, ?, ?, ?, ?, ?, ?)
                    """,
                        (
                            tool_name,
                            args_json,
                            response_json,
                            datetime.datetime.now(),
                            self.current_file_path,
                            request_id,
                            client_id,
                        ),
                    )
                    conn.commit()
                    logger.debug({"msg": f"Insert completed successfully"})
            except Exception as e:
                logger.debug({"msg": f"DuckDB error: {str(e)}"})
        except Exception as e:
            logger.debug({"msg": f"Error logging tool usage: {str(e)}"})

    def register_tools(self):
        @self.mcp.tool()
        async def set_file(filepath: str) -> str:
            """
            Set the current file to work with.

            This is always the first step in the workflow. You must set a file
            before you can use other tools like read, select etc.
            """

            if not os.path.isfile(filepath):
                return f"Error: File not found at '{filepath}'"

            # Check if the file path matches any of the protected paths
            for pattern in self.protected_paths:
                pattern = pattern.strip()
                if not pattern:
                    continue
                # Check for absolute path match
                if filepath == pattern:
                    return f"Error: Access to '{filepath}' is denied due to PROTECTED_PATHS configuration"
                # Check for glob pattern match (e.g., *.env, .env*, etc.)
                if "*" in pattern:
                    # First try matching the full path
                    if fnmatch.fnmatch(filepath, pattern):
                        return f"Error: Access to '{filepath}' is denied due to PROTECTED_PATHS configuration (matches pattern '{pattern}')"

                    # Then try matching just the basename
                    basename = os.path.basename(filepath)
                    if fnmatch.fnmatch(basename, pattern):
                        return f"Error: Access to '{filepath}' is denied due to PROTECTED_PATHS configuration (matches pattern '{pattern}')"

            self.current_file_path = filepath
            return f"File set to: '{filepath}'"

        @self.mcp.tool()
        async def skim() -> Dict[str, Any]:
            """
            Read text from the current file, truncated to the first `SKIM_MAX_LINES` lines.

            Returns:
                dict: lines, total_lines, max_select_lines
            """
            if self.current_file_path is None:
                return {"error": "No file path is set. Use set_file first."}
            with open(self.current_file_path, "r", encoding="utf-8") as file:
                lines = file.readlines()

                formatted_lines = []
                max_lines_to_show = int(os.getenv("SKIM_MAX_LINES", "500"))
                lines_to_process = lines[:max_lines_to_show]

                for i, line in enumerate(lines_to_process, 1):
                    formatted_lines.append((i, line.rstrip()))

            result = {
                "lines": formatted_lines,
                "total_lines": len(lines),
                "max_select_lines": self.max_select_lines,
            }

            # Add hint if file was truncated
            if len(lines) > max_lines_to_show:
                result["truncated"] = True
                result["hint"] = (
                    f"File has {len(lines)} total lines. Only showing first {max_lines_to_show} lines. Use `read` to view specific line ranges or `find_line` to search for content in the remaining lines."
                )

            return result

        @self.mcp.tool()
        async def read(start: int, end: int) -> Dict[str, Any]:
            """
            Read lines from the current file from start line to end line, returning them in a dictionary
            like {"lines":[[1,"text on first line"],[2,"text on second line"]]}. This makes it easier to find the precise lines to select for editing.

            Args:
                start (int, optional): Start line number
                end (int, optional): End line number

            Returns:
                dict: lines, start_line, end_line
            """
            result = {}

            if self.current_file_path is None:
                return {"error": "No file path is set. Use set_file first."}

            try:
                with open(self.current_file_path, "r", encoding="utf-8") as file:
                    lines = file.readlines()

                if start < 1:
                    return {"error": "start must be at least 1"}
                if end > len(lines):
                    end = len(lines)
                if start > end:
                    return {
                        "error": f"{start=} cannot be greater than {end=}. {len(lines)=}"
                    }

                selected_lines = lines[start - 1 : end]

                formatted_lines = []
                for i, line in enumerate(selected_lines, start):
                    formatted_lines.append((i, line.rstrip()))

                result["lines"] = formatted_lines
                result["start_line"] = start
                result["end_line"] = end

                return result

            except Exception as e:
                return {"error": f"Error reading file: {str(e)}"}

        @self.mcp.tool()
        async def select(
            start: int,
            end: int,
        ) -> Dict[str, Any]:
            """
            Select lines from for subsequent overwrite operation.

            Args:
                start (int): Start line number (1-based)
                end (int): End line number (1-based)

            Returns:
                dict: status, lines, start, end, id, line_count, message
            """
            if self.current_file_path is None:
                return {"error": "No file path is set. Use set_file first."}

            try:
                with open(self.current_file_path, "r", encoding="utf-8") as file:
                    lines = file.readlines()

                if start < 1:
                    return {"error": "start must be at least 1."}

                if end > len(lines):
                    end = len(lines)

                if start > end:
                    return {"error": "start cannot be greater than end."}

                if end - start + 1 > self.max_select_lines:
                    return {
                        "error": f"Cannot select more than {self.max_select_lines} lines at once (attempted {end - start + 1} lines)."
                    }

                selected_lines = lines[start - 1 : end]
                text = "".join(selected_lines)

                current_id = calculate_id(text, start, end)

                self.selected_start = start
                self.selected_end = end
                self.selected_id = current_id

                # Convert selected lines to a list without line numbers
                lines_content = [line.rstrip() for line in selected_lines]

                result = {
                    "status": "success",
                    "lines": lines_content,
                    "start": start,
                    "end": end,
                    "id": current_id,
                    "line_count": len(selected_lines),
                    "message": f"Selected lines {start} to {end} for editing.",
                }

                return result

            except Exception as e:
                return {"error": f"Error selecting lines: {str(e)}"}

        @self.mcp.tool()
        async def overwrite(
            new_lines: dict,
        ) -> Dict[str, Any]:
            """
            Overwrite the selected lines with new text. Amount of new lines can differ from the original selection

            Args:
                new_lines (dict): Example: {"lines":["line one", "second line"]}

            Returns:
                dict: Diff preview showing the proposed changes, and any syntax errors for JS or Python

            """
            new_lines = new_lines.get("lines")
            if self.current_file_path is None:
                return {"error": "No file path is set. Use set_file first."}

            if (
                self.selected_start is None
                or self.selected_end is None
                or self.selected_id is None
            ):
                return {"error": "No selection has been made. Use select tool first."}

            try:
                with open(self.current_file_path, "r", encoding="utf-8") as file:
                    lines = file.readlines()
            except Exception as e:
                return {"error": f"Error reading file: {str(e)}"}

            start = self.selected_start
            end = self.selected_end
            id = self.selected_id

            current_content = "".join(lines[start - 1 : end])

            computed_id = calculate_id(current_content, start, end)

            if computed_id != id:
                return {
                    "error": "id verification failed. The content may have been modified since you last read it."
                }

            processed_new_lines = []
            for line in new_lines:
                if not line.endswith("\n"):
                    processed_new_lines.append(line + "\n")
                else:
                    processed_new_lines.append(line)

            if (
                processed_new_lines
                and end < len(lines)
                and not processed_new_lines[-1].endswith("\n")
            ):
                processed_new_lines[-1] += "\n"

            before = lines[: start - 1]
            after = lines[end:]
            modified_lines = before + processed_new_lines + after
            diff_result = generate_diff_preview(lines, modified_lines, start, end)
            error = None
            if self.current_file_path.endswith(".py"):
                full_content = "".join(modified_lines)
                try:
                    black.format_file_contents(
                        full_content,
                        fast=True,
                        mode=black.Mode(),
                    )
                except black.InvalidInput as e:
                    error = {
                        "error": f"Python syntax error: {str(e)}",
                        "diff_lines": diff_result,
                        "auto_cancel": self.fail_on_python_syntax_error,
                    }
                except Exception as e:
                    if not isinstance(e, NothingChanged):
                        error = {
                            "error": f"Black check raised {type(e)}: {str(e)}",
                            "diff_lines": diff_result,
                        }

            elif self.enable_js_syntax_check and self.current_file_path.endswith(
                (".jsx", ".js")
            ):
                with tempfile.NamedTemporaryFile(
                    mode="w", suffix=".jsx", delete=False
                ) as temp:
                    temp_path = temp.name
                    temp.writelines(modified_lines)

                try:
                    presets = (
                        ["@babel/preset-react"]
                        if self.current_file_path.endswith(".jsx")
                        else ["@babel/preset-env"]
                    )

                    cmd = [
                        "npx",
                        "babel",
                        "--presets",
                        ",".join(presets),
                        "--no-babelrc",
                        temp_path,
                        "--out-file",
                        "/dev/null",  # Output to nowhere, we just want to check syntax
                    ]

                    # Execute Babel to transform (which validates syntax)
                    process = subprocess.run(cmd, capture_output=True, text=True)

                    if process.returncode != 0:
                        error_output = process.stderr

                        filtered_lines = []
                        for line in error_output.split("\n"):
                            if "node_modules/@babel" not in line:
                                filtered_lines.append(line)

                        filtered_error = "\n".join(filtered_lines).strip()

                        if not filtered_error:
                            filtered_error = "JavaScript syntax error detected"

                        error = {
                            "error": f"JavaScript syntax error: {filtered_error}",
                            "diff_lines": diff_result,
                            "auto_cancel": self.fail_on_js_syntax_error,
                        }

                except Exception as e:
                    os.unlink(temp_path)
                    error = {
                        "error": f"Error checking JavaScript syntax: {str(e)}",
                        "diff_lines": diff_result,
                    }

                finally:
                    if os.path.exists(temp_path):
                        os.unlink(temp_path)

            self.pending_modified_lines = modified_lines
            self.pending_diff = diff_result

            result = {
                "status": "preview",
                "message": "Changes ready to apply. Use confirm() to apply or cancel() to discard.",
                "diff_lines": diff_result["diff_lines"],
                "start": start,
                "end": end,
            }
            if error:
                result.update(error)
                if error.get("auto_cancel", False):
                    self.pending_modified_lines = None
                    self.pending_diff = None
                    result["status"] = "auto_cancelled"
                    result["message"] = (
                        "Changes automatically cancelled due to syntax error. The lines are still selected."
                    )
                else:
                    result["message"] += (
                        " It looks like there is a syntax error, but you can choose to fix it in the subsequent edits."
                    )

            return result

        # Later on this tool should be shown conditionally, however, most clients don't support this functionality yet.
        @self.mcp.tool()
        async def confirm() -> Dict[str, Any]:
            """Confirm action"""
            if self.pending_modified_lines is None or self.pending_diff is None:
                return {"error": "No pending changes to apply. Use overwrite first."}

            try:
                with open(self.current_file_path, "w", encoding="utf-8") as file:
                    file.writelines(self.pending_modified_lines)

                result = {
                    "status": "success",
                    "message": f"Changes applied successfully.",
                }

                self.selected_start = None
                self.selected_end = None
                self.selected_id = None
                self.pending_modified_lines = None
                self.pending_diff = None

                return result
            except Exception as e:
                return {"error": f"Error writing to file: {str(e)}"}

        @self.mcp.tool()
        async def cancel() -> Dict[str, Any]:
            """
            Cancel action
            """
            if self.pending_modified_lines is None or self.pending_diff is None:
                return {"error": "No pending changes to discard. Use overwrite first."}

            self.pending_modified_lines = None
            self.pending_diff = None

            return {
                "status": "success",
                "message": "Action cancelled.",
            }

        @self.mcp.tool()
        async def delete_file() -> Dict[str, Any]:
            """
            Delete current file
            """

            if self.current_file_path is None:
                return {"error": "No file path is set. Use set_file first."}

            try:
                if not os.path.exists(self.current_file_path):
                    return {"error": f"File '{self.current_file_path}' does not exist."}

                os.remove(self.current_file_path)

                deleted_path = self.current_file_path

                self.current_file_path = None

                return {
                    "status": "success",
                    "message": f"File '{deleted_path}' was successfully deleted.",
                }
            except Exception as e:
                return {"error": f"Error deleting file: {str(e)}"}

        @self.mcp.tool()
        async def new_file(filepath: str) -> Dict[str, Any]:
            """
            Creates a new file.

            After creating new file, the first line is automatically selected for editing.
            Automatically creates parent directories if they don't exist.

            Args:
                filepath (str): Path of the new file
            Returns:
                dict: Status message with selection info

            """
            self.current_file_path = filepath

            if (
                os.path.exists(self.current_file_path)
                and os.path.getsize(self.current_file_path) > 0
            ):
                return {
                    "error": "Cannot create new file. Current file exists and is not empty."
                }

            try:
                # Create parent directories if they don't exist
                directory = os.path.dirname(self.current_file_path)
                if directory:
                    os.makedirs(directory, exist_ok=True)

                text = "# NEW_FILE - REMOVE THIS HEADER"
                with open(self.current_file_path, "w", encoding="utf-8") as file:
                    file.write(text)

                # Automatically select the first line for editing
                self.selected_start = 1
                self.selected_end = 1
                self.selected_id = calculate_id(text, 1, 1)

                result = {
                    "status": "success",
                    "text": text,
                    "current_file_path": self.current_file_path,
                    "id": self.selected_id,
                    "selected_start": self.selected_start,
                    "selected_end": self.selected_end,
                    "message": "File created successfully. First line is now selected for editing.",
                }

                return result
            except Exception as e:
                return {"error": f"Error creating file: {str(e)}"}

        @self.mcp.tool()
        async def find_line(
            search_text: str,
        ) -> Dict[str, Any]:
            """
            Find lines that match provided text in the current file.

            Args:
                search_text (str): Text to search for in the file

            Returns:
                dict: Matching lines with their line numbers, and full text
            """
            if self.current_file_path is None:
                return {"error": "No file path is set. Use set_file first."}

            try:
                with open(self.current_file_path, "r", encoding="utf-8") as file:
                    lines = file.readlines()

                matches = []
                for i, line in enumerate(lines, start=1):
                    if search_text in line:
                        matches.append([i, line])

                result = {
                    "status": "success",
                    "matches": matches,
                    "total_matches": len(matches),
                }

                return result

            except Exception as e:
                return {"error": f"Error searching file: {str(e)}"}

        @self.mcp.tool()
        async def listdir(dirpath: str) -> Dict[str, Any]:
            try:
                return {
                    "filenames": os.listdir(dirpath),
                    "path": dirpath,
                }
            except NotADirectoryError as e:
                return {
                    "error": "Specified path is not a directory.",
                    "path": dirpath,
                }
            except Exception as e:
                return {
                    "error": f"Unexpected error when listing the directory: {str(e)}"
                }

        @self.mcp.tool()
        async def find_function(
            function_name: str,
        ) -> Dict[str, Any]:
            """
            Find a function or method definition in a Python or JS/JSX file. Uses AST parsers.

            Args:
                function_name (str): Name of the function or method to find

            Returns:
                dict: function lines with their line numbers, start_line, and end_line
            """
            if self.current_file_path is None:
                return {"error": "No file path is set. Use set_file first."}

            # Check if the file is a supported type (Python or JavaScript/JSX)
            is_python = self.current_file_path.endswith(".py")
            is_javascript = self.current_file_path.endswith((".js", ".jsx"))

            if not (is_python or is_javascript):
                return {
                    "error": "This tool only works with Python (.py) or JavaScript/JSX (.js, .jsx) files."
                }

            try:
                with open(self.current_file_path, "r", encoding="utf-8") as file:
                    source_code = file.read()
                    lines = source_code.splitlines(True)  # Keep line endings

                # Process JavaScript/JSX files
                if is_javascript:
                    return self._find_js_function(function_name, source_code, lines)

                # For Python files, parse the source code to AST
                tree = ast.parse(source_code)

                # Find the function in the AST
                function_node = None
                class_node = None
                parent_function = None

                # Helper function to find a function or method node
                def find_node(node):
                    nonlocal function_node, class_node, parent_function
                    if isinstance(node, ast.FunctionDef) and node.name == function_name:
                        function_node = node
                        return True
                    # Check for methods in classes
                    elif isinstance(node, ast.ClassDef):
                        for item in node.body:
                            if (
                                isinstance(item, ast.FunctionDef)
                                and item.name == function_name
                            ):
                                function_node = item
                                class_node = node
                                return True
                    # Check for nested functions
                    elif isinstance(node, ast.FunctionDef):
                        for item in node.body:
                            # Find directly nested function definitions
                            if (
                                isinstance(item, ast.FunctionDef)
                                and item.name == function_name
                            ):
                                function_node = item
                                # Store parent function information
                                parent_function = node
                                return True
                    # Recursively search for nested functions/methods
                    for child in ast.iter_child_nodes(node):
                        if find_node(child):
                            return True
                    return False

                # Search for the function in the AST
                find_node(tree)

                if not function_node:
                    return {
                        "error": f"Function or method '{function_name}' not found in the file."
                    }

                # Get the line range for the function
                start_line = function_node.lineno
                end_line = 0

                # Find the end line by looking at tokens
                with open(self.current_file_path, "rb") as file:
                    tokens = list(tokenize.tokenize(file.readline))

                # Find the function definition token
                function_def_index = -1
                for i, token in enumerate(tokens):
                    if token.type == tokenize.NAME and token.string == function_name:
                        if (
                            i > 0
                            and tokens[i - 1].type == tokenize.NAME
                            and tokens[i - 1].string == "def"
                        ):
                            function_def_index = i
                            break

                if function_def_index == -1:
                    # Fallback - use AST to determine the end
                    # First, get the end_lineno from the function node itself
                    end_line = function_node.end_lineno or start_line
                    # Then walk through all nodes inside the function to find the deepest end_lineno
                    # This handles nested functions and statements properly
                    # Walk through all nodes inside the function to find the deepest end_lineno
                    # This handles nested functions and statements properly
                    for node in ast.walk(function_node):
                        if hasattr(node, "end_lineno") and node.end_lineno:
                            end_line = max(end_line, node.end_lineno)

                    # Specifically look for nested function definitions
                    # by checking for FunctionDef nodes within the function body
                    for node in ast.walk(function_node):
                        if (
                            isinstance(node, ast.FunctionDef)
                            and node is not function_node
                        ):
                            if hasattr(node, "end_lineno") and node.end_lineno:
                                end_line = max(end_line, node.end_lineno)
                else:
                    # Find the closing token of the function (either the next function/class at the same level or the end of file)
                    indent_level = tokens[function_def_index].start[
                        1
                    ]  # Get the indentation of the function
                    in_function = False
                    nested_level = 0
                    for token in tokens[function_def_index + 1 :]:
                        current_line = token.start[0]
                        if current_line > start_line:
                            # Start tracking when we're inside the function body
                            if not in_function and token.string == ":":
                                in_function = True
                                continue

                            # Track nested blocks by indentation
                            if in_function:
                                current_indent = token.start[1]
                                # Find a token at the same indentation level as the function definition
                                # but only if we're not in a nested block
                                if (
                                    current_indent <= indent_level
                                    and token.type == tokenize.NAME
                                    and token.string in ("def", "class")
                                    and nested_level == 0
                                ):
                                    end_line = current_line - 1
                                    break
                                # Track nested blocks
                                elif (
                                    current_indent > indent_level
                                    and token.type == tokenize.NAME
                                ):
                                    if token.string in ("def", "class"):
                                        nested_level += 1
                                    # Look for the end of nested blocks
                                elif (
                                    nested_level > 0 and current_indent <= indent_level
                                ):
                                    nested_level -= 1

                    # If we couldn't find the end, use the last line of the file
                    if end_line == 0:
                        end_line = len(lines)

                # Include decorators if present
                for decorator in function_node.decorator_list:
                    start_line = min(start_line, decorator.lineno)

                # Adjust for methods inside classes
                if class_node:
                    class_body_start = min(
                        item.lineno
                        for item in class_node.body
                        if hasattr(item, "lineno")
                    )
                    if function_node.lineno == class_body_start:
                        # If this is the first method, include the class definition
                        start_line = class_node.lineno

                # Normalize line numbers (1-based for API consistency)
                function_lines = lines[start_line - 1 : end_line]

                # Format the results similar to the read tool
                formatted_lines = []
                for i, line in enumerate(function_lines, start_line):
                    formatted_lines.append((i, line.rstrip()))

                result = {
                    "status": "success",
                    "lines": formatted_lines,
                    "start_line": start_line,
                    "end_line": end_line,
                }

                # Add parent function information if this is a nested function
                if parent_function:
                    result["is_nested"] = True
                    result["parent_function"] = parent_function.name

                return result

            except Exception as e:
                return {"error": f"Error finding function: {str(e)}"}

        @self.mcp.tool()
        async def set_python_path(path: str):
            """
            Set it before running tests so the project is correctly recognized
            """
            os.environ["PYTHONPATH"] = path

        @self.mcp.tool()
        async def run_tests(
            test_path: Optional[str] = None,
            test_name: Optional[str] = None,
            verbose: bool = False,
            collect_only: bool = False,
        ) -> Dict[str, Any]:
            """
            Run pytest tests using the specified Python virtual environment.

            Args:
                test_path (str, optional): Directory or file path containing tests to run
                test_name (str, optional): Specific test function/method to run
                verbose (bool, optional): Run tests in verbose mode
                collect_only (bool, optional): Only collect tests without executing them

            Returns:
                dict: Test execution results including returncode, output, and execution time
            """
            if self.python_venv is None:
                return {
                    "error": "No Python environment found. It needs to be set in the MCP config as environment variable called PYTHON_VENV."
                }
            # Build pytest arguments
            pytest_args = []

            # Add test path if specified
            if test_path:
                pytest_args.append(test_path)

            # Add specific test name if specified
            if test_name:
                pytest_args.append(f"-k {test_name}")

            # Add verbosity flag if specified
            if verbose:
                pytest_args.append("-v")

            # Add collect-only flag if specified
            if collect_only:
                pytest_args.append("--collect-only")

            # Run the tests
            return self._run_tests(pytest_args)

    def _find_js_function(
        self, function_name: str, source_code: str, lines: list
    ) -> Dict[str, Any]:
        """
        Helper method to find JavaScript/JSX function definitions using Babel AST parsing.

        Args:
            function_name (str): Name of the function to find
            source_code (str): Source code content
            lines (list): Source code split by lines with line endings preserved

        Returns:
            dict: Dictionary with function information
        """
        try:
            # First try using Babel for accurate parsing if it's available
            if self.enable_js_syntax_check:
                babel_result = self._find_js_function_babel(
                    function_name, source_code, lines
                )
                if babel_result and not babel_result.get("error"):
                    return babel_result

            # Fallback to regex approach if Babel parsing fails or is disabled
            # Pattern for named function declaration
            # Matches: function functionName(args) { body }
            # Also matches: async function functionName(args) { body }
            function_pattern = re.compile(
                r"(?:async\s+)?function\s+(?P<functionName>\w+)\s*\((?P<functionArguments>[^()]*)\)\s*{",
                re.MULTILINE,
            )

            # Pattern for arrow functions with explicit name
            # Matches: const functionName = (args) => { body } or const functionName = args => { body }
            # Also matches async variants: const functionName = async (args) => { body }
            # Also matches component inner functions: const innerFunction = async () => { ... }
            arrow_pattern = re.compile(
                r"(?:(?:const|let|var)\s+)?(?P<functionName>\w+)\s*=\s*(?:async\s+)?(?:\((?P<functionArguments>[^()]*)\)|(?P<singleArg>\w+))\s*=>\s*{",
                re.MULTILINE,
            )

            # Pattern for object method definitions
            # Matches: functionName(args) { body } in object literals or classes
            # Also matches: async functionName(args) { body }
            method_pattern = re.compile(
                r"(?:^|,|{)\s*(?:async\s+)?(?P<functionName>\w+)\s*\((?P<functionArguments>[^()]*)\)\s*{",
                re.MULTILINE,
            )

            # Pattern for React hooks like useCallback, useEffect, etc.
            # Matches: const functionName = useCallback(async () => { ... }, [deps])
            hook_pattern = re.compile(
                r"const\s+(?P<functionName>\w+)\s*=\s*use\w+\((?:async\s+)?\(?[^{]*\)?\s*=>[^{]*{",
                re.MULTILINE,
            )

            # Search for the function
            matches = []

            # Check all patterns
            for pattern in [
                function_pattern,
                arrow_pattern,
                method_pattern,
                hook_pattern,
            ]:
                for match in pattern.finditer(source_code):
                    if match.groupdict().get("functionName") == function_name:
                        matches.append(match)

            if not matches:
                return {"error": f"Function '{function_name}' not found in the file."}

            # Use the first match
            match = matches[0]
            start_pos = match.start()

            # Find the line number for the start
            start_line = 1
            pos = 0
            for i, line in enumerate(lines, 1):
                next_pos = pos + len(line)
                if pos <= start_pos < next_pos:
                    start_line = i
                    break
                pos = next_pos

            # Find the closing brace that matches the opening brace of the function
            # Count the number of opening and closing braces
            brace_count = 0
            end_pos = start_pos
            in_string = False
            string_delimiter = None
            escaped = False

            for i in range(start_pos, len(source_code)):
                char = source_code[i]

                # Handle strings to avoid counting braces inside strings
                if not escaped and char in ['"', "'", "`"]:
                    if not in_string:
                        in_string = True
                        string_delimiter = char
                    elif char == string_delimiter:
                        in_string = False

                # Check for escape character
                if char == "\\" and not escaped:
                    escaped = True
                    continue

                escaped = False

                # Only count braces outside of strings
                if not in_string:
                    if char == "{":
                        brace_count += 1
                    elif char == "}":
                        brace_count -= 1
                        if brace_count == 0:
                            end_pos = i + 1  # Include the closing brace
                            break

            # Find the end line number
            end_line = 1
            pos = 0
            for i, line in enumerate(lines, 1):
                next_pos = pos + len(line)
                if pos <= end_pos < next_pos:
                    end_line = i
                    break
                pos = next_pos

            # Extract the function lines
            function_lines = lines[start_line - 1 : end_line]

            # Format results like the read tool
            formatted_lines = []
            for i, line in enumerate(function_lines, start_line):
                formatted_lines.append((i, line.rstrip()))

            result = {
                "status": "success",
                "lines": formatted_lines,
                "start_line": start_line,
                "end_line": end_line,
            }

            return result

        except Exception as e:
            return {"error": f"Error finding JavaScript function: {str(e)}"}

    def _find_js_function_babel(
        self, function_name: str, source_code: str, lines: list
    ) -> Dict[str, Any]:
        """
        Use Babel to parse JavaScript/JSX code and find function definitions.

        This provides more accurate function location by using proper AST parsing
        rather than regex pattern matching.

        Args:
            function_name (str): Name of the function to find
            source_code (str): Source code content
            lines (list): Source code split by lines with line endings preserved

        Returns:
            dict: Dictionary with function information or None if Babel fails
        """
        try:
            # Create a temporary file with the source code
            with tempfile.NamedTemporaryFile(
                mode="w", suffix=".jsx", delete=False
            ) as temp:
                temp_path = temp.name
                temp.write(source_code)

            # Determine the appropriate Babel preset
            is_jsx = self.current_file_path.endswith(".jsx")
            presets = ["@babel/preset-react"] if is_jsx else ["@babel/preset-env"]

            # Use Babel to output the AST as JSON
            cmd = [
                "npx",
                "babel",
                "--presets",
                ",".join(presets),
                "--plugins",
                # Add the AST plugin that outputs function locations
                "babel-plugin-ast-function-metadata",
                "--no-babelrc",
                temp_path,
                "--out-file",
                "/dev/null",  # Output to nowhere, we just want the AST metadata
            ]

            # Execute Babel to get the AST with function locations
            process = subprocess.run(cmd, capture_output=True, text=True)

            # Clean up the temporary file
            try:
                os.unlink(temp_path)
            except:
                pass

            # If Babel execution failed, return None to fall back to regex
            if process.returncode != 0:
                return None

            # Parse the output to find location data
            output = process.stdout
            # Look for the JSON that has our function location data
            location_data = None
            import json

            try:
                # Extract the JSON output from Babel plugin
                # Format is typically like: FUNCTION_LOCATIONS: {... json data ...}
                match = re.search(r"FUNCTION_LOCATIONS:\s*({.*})", output, re.DOTALL)
                if match:
                    json_str = match.group(1)
                    locations = json.loads(json_str)
                    # Find our specific function
                    location_data = locations.get(function_name)
            except (json.JSONDecodeError, AttributeError) as e:
                return None

            if not location_data:
                return None

            # Get the line information from the location data
            start_line = location_data.get("start", {}).get("line", 0)
            end_line = location_data.get("end", {}).get("line", 0)

            if start_line <= 0 or end_line <= 0:
                return None

            # Extract the function lines
            function_lines = lines[start_line - 1 : end_line]

            # Format results like the read tool
            formatted_lines = []
            for i, line in enumerate(function_lines, start_line):
                formatted_lines.append((i, line.rstrip()))

            result = {
                "status": "success",
                "lines": formatted_lines,
                "start_line": start_line,
                "end_line": end_line,
                "parser": "babel",  # Flag that this was parsed with Babel
            }

            return result

        except Exception as e:
            # If anything goes wrong, return None to fall back to regex approach
            return None

    def _run_tests(self, pytest_args=None, python_venv=None):
        """
        Run pytest tests using the specified Python virtual environment.

        Args:
            pytest_args (list, optional): List of arguments to pass to pytest
            python_venv (str, optional): Path to Python executable in virtual environment
                                     If not provided, uses PYTHON_VENV environment variable

        Returns:
            dict: Test execution results including returncode, output, and execution time
        """
        try:
            # Determine the Python executable to use
            python_venv = (
                python_venv or self.python_venv
            )  # Use the class level python_venv

            # If no venv is specified, use the system Python
            python_cmd = python_venv or "python"

            # Build the command to run pytest
            cmd = [python_cmd, "-m", "pytest"]

            # Add any additional pytest arguments
            if pytest_args:
                cmd.extend(pytest_args)

            # Record the start time
            start_time = datetime.datetime.now()

            # Run the pytest command
            process = subprocess.run(cmd, capture_output=True, text=True)

            # Record the end time and calculate duration
            end_time = datetime.datetime.now()
            duration = (end_time - start_time).total_seconds()

            # Return the results
            return {
                "status": "success" if process.returncode == 0 else "failure",
                "returncode": process.returncode,
                "stdout": process.stdout,
                "stderr": process.stderr,
                "duration": duration,
                "command": " ".join(cmd),
            }
        except Exception as e:
            return {
                "status": "error",
                "error": str(e),
                "command": " ".join(cmd) if "cmd" in locals() else None,
            }

    def run(self, transport="stdio", **transport_kwargs):
        """Run the MCP server."""
        self.mcp.run(transport=transport, **transport_kwargs)


def main():
    """Entry point for the application.

    This function is used both for direct execution and
    when the package is installed via UVX, allowing the
    application to be run using the `editor-mcp` command.
    """

    parser = argparse.ArgumentParser(description="Text Editor MCP Server")
    parser.add_argument(
        "--transport",
        default="stdio",
        choices=["stdio", "http", "streamable-http"],
        help="Transport type",
    )
    parser.add_argument(
        "--host", default="127.0.0.1", help="Host to bind to (for HTTP transport)"
    )
    parser.add_argument(
        "--port", type=int, default=8001, help="Port to bind to (for HTTP transport)"
    )
    parser.add_argument(
        "--path", default="/mcp", help="Path for HTTP endpoint (for HTTP transport)"
    )

    args = parser.parse_args()

    host = os.environ.get("FASTMCP_SERVER_HOST", args.host)
    port = int(os.environ.get("FASTMCP_SERVER_PORT", args.port))
    path = os.environ.get("FASTMCP_SERVER_PATH", args.path)
    transport = os.environ.get("FASTMCP_SERVER_TRANSPORT", args.transport)

    # Normalize transport name for FastMCP
    if transport in ["http", "streamable-http"]:
        transport = "streamable-http"

    # Run the server with the configured transport
    if transport == "streamable-http":
        text_editor_server.run(transport=transport, host=host, port=port, path=path)
    else:
        text_editor_server.run(transport=transport)


text_editor_server = TextEditorServer()
mcp = text_editor_server.mcp
if __name__ == "__main__":
    main()
